using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.DigitalAudio
{
    public class WAVFile16 : BLDA.WAVFile
    {
        //an array to store the PCM data
        protected short[][] _Channels;

        #region Properties

        public short[][] Channels
        {
            get
            {
                return _Channels;
            }
            set
            {
                _Channels = value;
            }
        }

        #endregion

        public WAVFile16(ushort numChannels, uint sampleRate, ushort bitsPerSample, uint numSamplesPerChannel)
            : base(numChannels, sampleRate, bitsPerSample, numSamplesPerChannel)
        {
        }

        public override void InitializeChannels()
        {
            _Channels = new short[_NumChannels][];
            for (int channelNum = 0; channelNum < NumChannels; channelNum++)
            {
                _Channels[channelNum] = new short[_NumSamplesPerChannel];
                for (uint sampleNum = 0; sampleNum < _NumSamplesPerChannel; sampleNum++)
                {
                    _Channels[channelNum][sampleNum] = 0;
                }
            }
        }

        public override void WriteData(Stream outStream)
        {
            BinaryWriter oWriter = new BinaryWriter(outStream);
            for (int sampleNum = 0; sampleNum < _NumSamplesPerChannel; sampleNum++)
            {
                for (int channelNum = 0; channelNum < _NumChannels; channelNum++)
                {
                    oWriter.Write(_Channels[channelNum][sampleNum]);
                }
            }
            oWriter.Close();
        }

        public override void ReadData(Stream inStream)
        {
            BinaryReader iReader = new BinaryReader(inStream);
            int lastSampleRead = 0;
            try
            {
                for (int sampleNum = 0; sampleNum < _NumSamplesPerChannel; sampleNum++)
                {
                    for (int channelNum = 0; channelNum < _NumChannels; channelNum++)
                    {
                        _Channels[channelNum][sampleNum] = iReader.ReadInt16();
                        lastSampleRead++;
                    }
                }
            }
            catch (EndOfStreamException eosEx)
            {
                throw new BLS.UnexpectedEndOfStreamException("Reached end of stream after reading " + lastSampleRead.ToString() + " samples. " +
                    "Expected " + _NumSamplesPerChannel.ToString() + " samples total.");
            }
            finally
            {
                iReader.Close();
            }
        }
    }
}
