// BenLincoln.HotKeys
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.HotKeys.

// BenLincoln.HotKeys is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.HotKeys is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.HotKeys (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.HotKeys
{
    public class HotKey : IMessageFilter
    {
        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool RegisterHotKey(IntPtr hWnd, int id, KeyModifiers fsModifiers, Keys vk);
        [DllImport("user32.dll", SetLastError = true)]
        private static extern bool UnregisterHotKey(IntPtr hWnd, int id);

        public enum KeyModifiers
        {
            None = 0,
            Alt = 1,
            Control = 2,
            Shift = 4,
            Windows = 8
        }

        protected const int WM_HOTKEY = 0x0312;

        protected IntPtr _Handle;
        protected int _ID;
        protected event EventHandler _OnHotKey;

        #region Properties

        public IntPtr Handle
        {
            get
            {
                return _Handle;
            }
            set
            {
                _Handle = value;
            }
        }

        #endregion

        public HotKey(IntPtr handle, int id, Keys key, KeyModifiers keyModifier, EventHandler onHotKey)
        {
            _Handle = handle;
            _ID = id;
            _OnHotKey = onHotKey;
            RegisterHotKey(key, keyModifier);
            Application.AddMessageFilter(this);
        }

        ~HotKey()
        {
            Application.RemoveMessageFilter(this);
            UnregisterHotKey(_Handle, _ID);
        }

        protected void RegisterHotKey(Keys key, KeyModifiers keyModifier)
        {
            if (key == Keys.None)
            {
                throw new HotKeyException("A hotkey can only be bound to a key being pressed.");
            }
            bool result = RegisterHotKey(_Handle, _ID, keyModifier, key);
            if (result == false)
            {
                throw new HotKeyException("Hotkey is already registered.");
            }
        }

        public bool PreFilterMessage(ref Message m)
        {
            if ((m.Msg == WM_HOTKEY) && (m.WParam.ToInt32() == _ID))
            {
                _OnHotKey(this, new EventArgs());
                return true;
            }
            return false;
        }

    }
}
