// BenLincoln.Imaging
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.Imaging.

// BenLincoln.Imaging is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.Imaging is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.Imaging (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;

namespace BenLincoln.Imaging
{
    public abstract class ColourPalette
    {
        public static ColorPalette FromArray(Color[] colours)
        {
            int numColours = colours.GetUpperBound(0) + 1;
            ColorPalette pal = CreatePalette(numColours);
            for (int i = 0; i < numColours; i++)
            {
                pal.Entries[i] = colours[i];
            }
            return pal;
        }

        protected static ColorPalette CreatePalette(int numColours)
        {
            // based on MS sample code
            // http://support.microsoft.com/kb/319061
            PixelFormat format = PixelFormat.Format8bppIndexed;

            switch (numColours)
            {
                case 16:
                    format = PixelFormat.Format4bppIndexed;
                    break;
                case 256:
                    format = PixelFormat.Format8bppIndexed;
                    break;
                default:
                    throw new NotImplementedException("Only 16- and 256-colour palettes can be created by this method.");
                    break;
            }

            Bitmap image = new Bitmap(1, 1, format);
            ColorPalette pal = image.Palette;
            image.Dispose();

            return pal;
        }
    }
}
