// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class SoundClipContainerFile : IDisposable, BLS.IProgress
    {
        // defined in subclasses
        protected string _TypeName;

        // path to the file
        protected string _Path;

        // clips in the file
        protected BLCDDA.SoundClip[] _SoundClips;

        // loading progress
        protected double _IOPercent = 0;

        #region Properties

        public string TypeName
        {
            get
            {
                return _TypeName;
            }
        }

        public string Path
        {
            get
            {
                return _Path;
            }
        }

        public BLCDDA.SoundClip[] SoundClips
        {
            get
            {
                return _SoundClips;
            }
        }

        #endregion

        public SoundClipContainerFile(string path)
        {
            _Path = path;
        }

        public virtual void LoadFile()
        {
        }

        public void LoadClipData()
        {
            int clipCount = _SoundClips.GetUpperBound(0) + 1;
            double clipFactor = (double)1 / (double)clipCount;
            for (int clipNum = 0; clipNum < clipCount; clipNum++)
            {
                Thread loadThread = new Thread(_SoundClips[clipNum].LoadFileData);
                loadThread.Start();
                // wait for the loading to finish, and keep the load value in this class updated
                do
                {
                    _IOPercent = (clipFactor * (double)clipNum) + (clipFactor * _SoundClips[clipNum].Progress());
                    Thread.Sleep(5);
                }
                while (loadThread.IsAlive);
            }
        }

        // get a hashtable of all the global properties of the file
        public Hashtable GetInfo()
        {
            Hashtable info = GetSubclassInfo();
            info.Add("Path", _Path);
            if (_SoundClips != null)
            {
                info.Add("Number of Clips", _SoundClips.GetLength(0));
            } 
            return info;
        }

        // subclasses only have to define things that are unique to them
        protected virtual Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "Sound Clip Container File Superclass");
            return info;
        }

        public double Progress()
        {
            return _IOPercent;
        }

        public void Dispose()
        {
            Dispose(true);
        }

        protected void Dispose(bool disposing)
        {
            if (disposing && (_SoundClips != null))
            {
                foreach (BLCDDA.SoundClip clip in _SoundClips)
                {
                    if (clip != null)
                    {
                        clip.Dispose();
                    }
                }
            }
        }
    }
}
