// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class ERGVariable
    {
        protected string _Category;
        protected string _Name;
        protected Control[] _BoundControls;

        #region Properties

        public string Category
        {
            get
            {
                return _Category;
            }
        }

        public string Name
        {
            get
            {
                return _Name;
            }
        }

        #endregion

        public ERGVariable(string category, string name, Control boundControl)
        {
            _Category = category;
            _Name = name;
            BindControls(new Control[] { boundControl });
        }

        public virtual void BindControls(Control[] boundControls)
        {
            _BoundControls = boundControls;
            UpdateBoundControls();
        }

        public string GetFullName()
        {
            StringBuilder builder = new StringBuilder();
            if (_Category != "")
            {
                builder.Append(_Category + "\\");
            }
            builder.Append(_Name);
            string fullName = builder.ToString();
            Regex specialChars = new Regex("[^a-zA-Z0-9]");
            if (specialChars.Match(fullName).Success)
            {
                fullName = "\"" + fullName + "\"";
            }
            return fullName;
        }

        public virtual string GetValue()
        {
            return "null";
        }

        public void SetValue(string newValue)
        {
            SetValue(newValue, false);
        }

        public virtual void SetValue(string newValue, bool invertValue)
        {
            // overridden in subclasses
            // invertvalue is for handling the "-somevalue" syntax in the default file
        }

        public virtual void UpdateBoundControls()
        {
            // overridden in subclasses
            // changes the state of the bound control to match the state of the variable
        }

        public virtual void GetValueFromBoundControls()
        {
            // overridden in subclasses
            // changes the state of the variable to match the state of the bound control
        }
    }
}
