// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class ERGVariableStartArea : ERGVariable
    {
        protected string _LevelName;
        protected string _AreaName;
        protected GameArea _Area;
        protected GameLevelCollection _Levels;

        #region Properties

        public string LevelName
        {
            get
            {
                return _LevelName;
            }
        }

        public string AreaName
        {
            get
            {
                return _AreaName;
            }
        }

        #endregion

        public ERGVariableStartArea(string category, string name,
            Control boundLevelControl, Control boundAreaControl, Control boundDescriptionControl,
            string initialValue, GameLevelCollection levels)
            : base(category, name, null)
        {
            BindControls(new Control[] { boundLevelControl, boundAreaControl, boundDescriptionControl });
            _Levels = levels;
            _Area = _Levels.GetAreaByFullPath(initialValue);
        }

        public override void BindControls(Control[] boundControls)
        {
            _BoundControls = boundControls;
            if (_BoundControls[0] != null)
            {
                try
                {
                    ComboBox bound = GetBoundControlAsComboBox(0);
                    //bound.SelectedIndexChanged += new EventHandler(BoundControlHandler);
                    // not bound because an action should only be taken when the area is selected
                }
                catch (InvalidCastException cEx)
                {
                    throw new InvalidCastException("Start Level can only be bound to a control of type ComboBox.");
                }
            }
            if ((_BoundControls.GetUpperBound(0) > 0) && (_BoundControls[1] != null))
            {
                try
                {
                    ComboBox bound = GetBoundControlAsComboBox(1);
                    bound.SelectedIndexChanged += new EventHandler(BoundControlHandler);
                }
                catch (InvalidCastException cEx)
                {
                    throw new InvalidCastException("Start Area can only be bound to a control of type ComboBox.");
                }
            }
            if ((_BoundControls.GetUpperBound(0) > 1) && (_BoundControls[2] != null))
            {
                try
                {
                    Label bound = GetBoundControlAsLabel();
                }
                catch (InvalidCastException cEx)
                {
                    throw new InvalidCastException("Start Area Description can only be bound to a control of type Label.");
                }
            }
            UpdateBoundControls();
        }

        public void BoundControlHandler(object sender, EventArgs e)
        {
            GetValueFromBoundControls();
        }

        public ComboBox GetBoundControlAsComboBox(int controlNumber)
        {
            if (controlNumber > 1)
            {
                throw new InvalidCastException("Only controls 0 and 1 of a Start Area variable can be bound to a ComboBox.");
            }
            ComboBox bound;
            try
            {
                bound = (ComboBox)_BoundControls[controlNumber];
            }
            catch (InvalidCastException cEx)
            {
                throw new InvalidCastException("Bound controls 0 and 1 of a Start Area variable must be a control of type ComboBox.");
            }
            return bound;
        }

        public Label GetBoundControlAsLabel()
        {
            Label bound;
            try
            {
                bound = (Label)_BoundControls[2];
            }
            catch (InvalidCastException cEx)
            {
                throw new InvalidCastException("Bound control 2 of a Start Area variable must be a control of type Label.");
            }
            return bound;
        }

        public override string GetValue()
        {
            if (_Area == null)
            {
                return "";
            }
            else
            {
                return "\"" + _Area.GetFullPath() + "\"";
            }
        }

        protected GameArea GetArea()
        {
            if (_Levels.ContainsLevel(_LevelName))
            {
                GameLevel level = _Levels.GetLevelByName(_LevelName);
                if (level.ContainsArea(_AreaName))
                {
                    return level.GetAreaByName(_AreaName);
                }
            }
            return null;
        }

        //public static string[] GetLevelAndArea(string value)
        //{
        //    // get rid of quotes on values read from game.erg
        //    string nVal = value;
        //    if (nVal.StartsWith("\""))
        //    {
        //        nVal = nVal.Substring(1);
        //    }
        //    if (nVal.EndsWith("\""))
        //    {
        //        nVal = nVal.Substring(0, nVal.Length - 1);
        //    }
        //    // split out the area name by using the last forward slash
        //    int lastSlashIndex = nVal.LastIndexOf('/');
        //    string[] components = new string[2];
        //    components[0] = nVal.Substring(lastSlashIndex + 1);
        //    components[1] = nVal.Substring(0, lastSlashIndex);
        //    return components;
        //}

        public override void SetValue(string newValue, bool invertValue)
        {
            //string[] names = GetLevelAndArea(newValue);
            _Area = _Levels.GetAreaByFullPath(newValue);
            _LevelName = _Area.ParentLevel.Name;
            _AreaName = _Area.Name;
            UpdateBoundControls();
        }

        public override void UpdateBoundControls()
        {
            string lastAreaName = _AreaName;
            if (_BoundControls[0] != null)
            {
                SetComboBoxIndex(0, _LevelName);
            }
            if ((_BoundControls.GetUpperBound(0) > 0) && (_BoundControls[1] != null))
            {
                SetComboBoxIndex(1, lastAreaName);
            }
            if ((_BoundControls.GetUpperBound(0) > 1) && (_BoundControls[2] != null))
            {
                Label bound = GetBoundControlAsLabel();
                bound.Text = "Description: " + _Area.Description + ".";
            }
        }

        protected void SetComboBoxIndex(int controlNum, string selectedValue)
        {
            if (selectedValue == null)
            {
                return;
            }
            ComboBox bound = GetBoundControlAsComboBox(controlNum);
            if (bound.Items.Contains(selectedValue))
            {
                bound.SelectedItem = selectedValue;
            }
            else
            {
                bound.SelectedIndex = 0;
            }
        }

        public override void GetValueFromBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                ComboBox bound = GetBoundControlAsComboBox(0);
                _LevelName = bound.SelectedItem.ToString();
            }
            if (_BoundControls[1] != null)
            {
                ComboBox bound = GetBoundControlAsComboBox(1);
                _AreaName = bound.SelectedItem.ToString();
            }
            _Area = GetArea();
            if ((_BoundControls.GetUpperBound(0) > 1) && (_BoundControls[2] != null))
            {
                Label bound = GetBoundControlAsLabel();
                bound.Text = "Description: " + _Area.Description + ".";
            }
        }

    }
}
