// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class ERGVariableString : ERGVariable
    {
        protected string _Value;

        #region Properties

        public string Value
        {
            get
            {
                return _Value;
            }
        }

        #endregion

        public ERGVariableString(string category, string name, Control boundControl, string initialValue)
            : base(category, name, boundControl)
        {
            if (boundControl != null)
            {
                try
                {
                    TextBox bound = GetBoundControlAsTextBox();
                    bound.TextChanged += new EventHandler(BoundControlHandler);
                }
                catch (InvalidCastException cEx)
                {
                    throw new InvalidCastException("String variables can only be bound to a control of type TextBox.");
                }
            }
            _Value = initialValue.Trim();
        }

        public override void BindControls(Control[] boundControls)
        {
            _BoundControls = boundControls;
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox();
                bound.TextChanged += new EventHandler(BoundControlHandler);
            }
            UpdateBoundControls();
        }

        public void BoundControlHandler(object sender, EventArgs e)
        {
            GetValueFromBoundControls();
        }

        public TextBox GetBoundControlAsTextBox()
        {
            TextBox bound;
            try
            {
                bound = (TextBox)_BoundControls[0];
            }
            catch (InvalidCastException cEx)
            {
                throw new InvalidCastException("String variables can only be bound to a control of type TextBox.");
            }
            return bound;
        }

        public override string GetValue()
        {
            return "\"" + _Value + "\"";
        }

        public override void SetValue(string newValue, bool invertValue)
        {
            // get rid of quotes on values read from game.erg
            string nVal = newValue.Trim();
            if (nVal.StartsWith("\""))
            {
                nVal = nVal.Substring(1);
            }
            if (nVal.EndsWith("\""))
            {
                nVal = nVal.Substring(0, nVal.Length - 1);
            }
            _Value = nVal;
            UpdateBoundControls();
        }

        public override void UpdateBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox();
                bound.Text = _Value;
            }
        }

        public override void GetValueFromBoundControls()
        {
            if (_BoundControls[0] != null)
            {
                TextBox bound = GetBoundControlAsTextBox();
                SetValue(bound.Text);
            }
        }
    }
}
