// BenLincoln.TheLostWorlds.CDGameDotErg
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDGameDotErg.

// BenLincoln.TheLostWorlds.CDGameDotErg is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDGameDotErg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDGameDotErg (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDGameDotErg
{
    public class GameLevelCollection
    {
        protected GameLevel[] _Levels;

        #region Properties

        public GameLevel[] Levels
        {
            get
            {
                return _Levels;
            }
            set
            {
                _Levels = value;
            }
        }

        #endregion

        public GameLevelCollection(int numLevels)
        {
            _Levels = new GameLevel[numLevels];
        }

        public GameLevel GetLevelByName(string name)
        {
            foreach (GameLevel level in _Levels)
            {
                if (level.Name == name)
                {
                    return level;
                }
            }
            throw new KeyNotFoundException("Level collection does not contain a level called '" + name + "'.");
        }

        public bool ContainsLevel(string name)
        {
            foreach (GameLevel level in _Levels)
            {
                if (level.Name == name)
                {
                    return true;
                }
            }
            return false;
        }

        public GameArea GetAreaByFullPath(string fullPath)
        {
            foreach (GameLevel level in _Levels)
            {
                foreach (GameArea area in level.Areas)
                {
                    if (area.GetFullPath().ToLower() == fullPath.ToLower())
                    {
                        return area;
                    }
                }
            }
            throw new KeyNotFoundException("Level collection does not contain an area with the full path '" + fullPath + "'.");
        }

        public bool ContainsAreaByFullPath(string fullPath)
        {
            foreach (GameLevel level in _Levels)
            {
                foreach (GameArea area in level.Areas)
                {
                    if (area.GetFullPath().ToLower() == fullPath.ToLower())
                    {
                        return true;
                    }
                }
            }
            return false;
        }
    }
}
