// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

// This class and its subclasses are a big mess
// When I was going through the source code to this application to clean it up for release,
// I discovered that I had apparently abandoned it midway through refactoring it, and hadn't saved
// The source code from the last working version. Or the last truly working version, apparently.
// The code here works, but it is not pretty. Sorry.
// - Ben Lincoln, 21 January 2012

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDImages
{
    #region Enums

    public enum RAWHeaderSignature
    {
        Unknown,
        Standard,
        SR2PS2AirForgeDemo
    }

    #endregion

    #region Structs

    public struct RAWParameters
    {
        public int Width;
        public int Height;

        public long DataLength;
        public int Stride;

        public int BPPImage;

        public int BPP1;
        public int BPP2;
        public int BPP3;
        public int BPP4;

        public uint Mask1;
        public uint Mask2;
        public uint Mask3;
        public uint Mask4;

        public ushort StartBitAlpha;
        public ushort StartBitRed;
        public ushort StartBitGreen;
        public ushort StartBitBlue;

        public TypeFlags TypeFlags;
    }

    public struct TypeFlags
    {
        public uint Flags1; // FileType in old classes
        public uint Flags2; // ImageType in old classes
        public uint Flags3; // unknown3 in old classes
        public uint Flags4; // SubWidths[0] in old classes
    }

    #endregion

    public abstract class RAWImage
    {
        protected const int _HeaderOffset = 8;
        protected const int _MinimumFileSize = 128;

        protected string _FilePath;
        protected string _TypeDescription;
        protected string _FormatDescription;

        protected long _DataOffset;

        protected RAWParameters _RAWParameters;

        protected int _BPPAlpha;
        protected int _BPPRed;
        protected int _BPPGreen;
        protected int _BPPBlue;

        protected uint _MaskAlpha;
        protected uint _MaskRed;
        protected uint _MaskGreen;
        protected uint _MaskBlue;

        protected bool _IgnoreAlpha;

        #region Properties

        public string FilePath
        {
            get
            {
                return _FilePath;
            }
        }

        public string TypeDescription
        {
            get
            {
                return _TypeDescription;
            }
        }

        public string FormatDescription
        {
            get
            {
                return _FormatDescription;
            }
        }

        public long DataOffset
        {
            get
            {
                return _DataOffset;
            }
        }

        public int Width
        {
            get
            {
                return _RAWParameters.Width;
            }
        }

        public int Height
        {
            get
            {
                return _RAWParameters.Height;
            }
        }

        public long DataLength
        {
            get
            {
                return _RAWParameters.DataLength;
            }
        }

        public int Stride
        {
            get
            {
                return _RAWParameters.Stride;
            }
        }

        public int BPPImage
        {
            get
            {
                return _RAWParameters.BPPImage;
            }
        }

        public int BPPAlpha
        {
            get
            {
                return _BPPAlpha;
            }
            set
            {
                _BPPAlpha = value;
            }
        }

        public int BPPRed
        {
            get
            {
                return _BPPRed;
            }
            set
            {
                _BPPRed = value;
            }
        }

        public int BPPGreen
        {
            get
            {
                return _BPPGreen;
            }
            set
            {
                _BPPGreen = value;
            }
        }

        public int BPPBlue
        {
            get
            {
                return _BPPBlue;
            }
            set
            {
                _BPPBlue = value;
            }
        }

        public uint MaskAlpha
        {
            get
            {
                return _MaskAlpha;
            }
            set
            {
                _MaskAlpha = value;
            }
        }

        public uint MaskRed
        {
            get
            {
                return _MaskRed;
            }
            set
            {
                _MaskRed = value;
            }
        }

        public uint MaskGreen
        {
            get
            {
                return _MaskGreen;
            }
            set
            {
                _MaskGreen = value;
            }
        }

        public uint MaskBlue
        {
            get
            {
                return _MaskBlue;
            }
            set
            {
                _MaskBlue = value;
            }
        }

        public ushort StartBitAlpha
        {
            get
            {
                return _RAWParameters.StartBitAlpha;
            }
        }

        public ushort StartBitRed
        {
            get
            {
                return _RAWParameters.StartBitRed;
            }
        }

        public ushort StartBitGreen
        {
            get
            {
                return _RAWParameters.StartBitGreen;
            }
        }

        public ushort StartBitBlue
        {
            get
            {
                return _RAWParameters.StartBitBlue;
            }
        }

        public bool IgnoreAlpha
        {
            get
            {
                return _IgnoreAlpha;
            }
            set
            {
                _IgnoreAlpha = value;
            }
        }

        #endregion

        public abstract Bitmap GetBitmap();

        public virtual string GetInfo()
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("Path: ");
            builder.Append(FilePath);
            builder.Append(Environment.NewLine);

            builder.Append("Type: ");
            builder.Append(TypeDescription);
            builder.Append(Environment.NewLine);

            builder.Append("Format: ");
            builder.Append(FormatDescription);
            builder.Append(Environment.NewLine);

            builder.Append("Width: ");
            builder.Append(Width.ToString());
            builder.Append(" pixels");
            builder.Append(Environment.NewLine);

            builder.Append("Height: ");
            builder.Append(Height.ToString());
            builder.Append(" pixels");
            builder.Append(Environment.NewLine);

            builder.Append("Bits Per Pixel (Total): ");
            builder.Append(string.Format("0:00", BPPImage));
            builder.Append(Environment.NewLine);

            builder.Append("Bits Per Pixel (Alpha): ");
            builder.Append(string.Format("0:00", BPPAlpha));
            builder.Append(" / Mask: 0x");
            builder.Append(string.Format("0:X8", MaskAlpha));
            builder.Append(Environment.NewLine);

            builder.Append("Bits Per Pixel (Red): ");
            builder.Append(string.Format("0:00", BPPRed));
            builder.Append(" / Mask: 0x");
            builder.Append(string.Format("0:X8", MaskRed));
            builder.Append(Environment.NewLine);

            builder.Append("Bits Per Pixel (Green): ");
            builder.Append(string.Format("0:00", BPPGreen));
            builder.Append(" / Mask: 0x");
            builder.Append(string.Format("0:X8", MaskGreen));
            builder.Append(Environment.NewLine);

            builder.Append("Bits Per Pixel (Blue): ");
            builder.Append(string.Format("0:00", BPPBlue));
            builder.Append(" / Mask: 0x");
            builder.Append(string.Format("0:X8", MaskBlue));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        public static RAWParameters ReadRAWParameters(string path)
        {
            RAWParameters parms = new RAWParameters();

            FileStream iStream;
            BinaryReader iReader;
            iStream = new FileStream(path, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);
            iStream.Seek(_HeaderOffset, SeekOrigin.Begin);

            parms.DataLength = iReader.ReadUInt32();
            parms.TypeFlags.Flags1 = iReader.ReadUInt32();
            parms.TypeFlags.Flags2 = iReader.ReadUInt32();
            parms.Width = (int)iReader.ReadUInt32();
            parms.Height = (int)iReader.ReadUInt32();
            parms.Stride = (int)iReader.ReadUInt32();
            parms.BPPImage = (int)iReader.ReadUInt32();
            parms.BPP1 = (int)iReader.ReadUInt32();
            parms.BPP2 = (int)iReader.ReadUInt32();
            parms.BPP3 = (int)iReader.ReadUInt32();
            parms.BPP4 = (int)iReader.ReadUInt32();
            parms.Mask1 = iReader.ReadUInt32();
            parms.Mask2 = iReader.ReadUInt32();
            parms.Mask3 = iReader.ReadUInt32();
            parms.Mask4 = iReader.ReadUInt32();
            uint tUnknown1 = iReader.ReadUInt32();
            uint tUnknown2 = iReader.ReadUInt32();
            parms.TypeFlags.Flags3 = iReader.ReadUInt32();
            uint tUnknown4 = iReader.ReadUInt32();
            parms.TypeFlags.Flags4 = iReader.ReadUInt32();

            iReader.Close();
            iStream.Close();

            return parms;
        }

        protected void UpdateStartBits()
        {
            //transform the parameters now that we have all the data
            if (BPPImage == 32)
            {
                _RAWParameters.StartBitAlpha = (ushort)getBitPositionInDWord(MaskAlpha);
                _RAWParameters.StartBitRed = (ushort)getBitPositionInDWord(MaskRed);
                _RAWParameters.StartBitGreen = (ushort)getBitPositionInDWord(MaskGreen);
                _RAWParameters.StartBitBlue = (ushort)getBitPositionInDWord(MaskBlue);
            }
            else if (BPPImage == 16)
            {
                _RAWParameters.StartBitAlpha = (ushort)getBitPositionInWord(MaskAlpha);
                _RAWParameters.StartBitRed = (ushort)getBitPositionInWord(MaskRed);
                _RAWParameters.StartBitGreen = (ushort)getBitPositionInWord(MaskGreen);
                _RAWParameters.StartBitBlue = (ushort)getBitPositionInWord(MaskBlue);
            }
        }

        public static RAWImage FromFile(string path)
        {
            if (!File.Exists(path))
            {
                throw new FileNotFoundException("The file '" + path + "' does not exist.");
            }
            try
            {
                FileInfo info = new FileInfo(path);
                if (info.Length < _MinimumFileSize)
                {
                    throw new RAWImageException("The file '" + path + "' is not long enough to be a RAW image.");
                }
            }
            catch (Exception ex)
            {
                throw new RAWImageException("An error occurred while accessing the file '" + path + "'.", ex);
            }

            // get header and determine if it matches a known signature
            RAWHeaderSignature signature = GetHeaderData(path);

            // code to determine file type and return an instance of that sub class
            RAWParameters parms;
            try
            {
                parms = ReadRAWParameters(path);
            }
            catch (EndOfStreamException eoEx)
            {
                throw new RAWImageException("The file '" + path + "' is not long enough to be a RAW image.", eoEx);
            }
            catch (Exception ex)
            {
                throw new RAWImageException("Unable to read the file's parameters.", ex);
            }

            switch (signature)
            {
                case RAWHeaderSignature.Unknown:
                    throw new RAWImageException("The file '" + path + "' is not a RAW image, or is in an unsupported format.");
                    break;
                case RAWHeaderSignature.SR2PS2AirForgeDemo:
                    if ((parms.TypeFlags.Flags1 == 0x04) && (parms.TypeFlags.Flags2 == 0x400000))
                    {
                        return new SR2PS2AirForgeIndexedRAWImage(path, parms);
                    }
                    else
                    {
                        return new SR2PS2AirForgeARGBRAWImage(path, parms); 
                    }
                    break;
                default:
                    // take no action, additional processing is required
                    break;
            }

            // this data is only the segmentation/tiling parameters in that type of file, but 
            // is read here to simplify the code below
            SegmentedRAWParameters segParms = SegmentedRAWImage.GetSegmentedParameters(path);

            // the order in which these statements are placed is very important
            // because there is no universal way of determining the file type
            if ((parms.TypeFlags.Flags1 == 0x10) && (parms.TypeFlags.Flags2 == 0x10))
            {
                return new DefiancePCRAWImage(path, parms);
            }
            if ((parms.TypeFlags.Flags1 == 0x04) && (parms.TypeFlags.Flags2 == 0x400000))
            {
                return new SR2PS2IndexedRAWImage(path, parms, segParms);
            }
            // it would be nice if there were a more elegant way to do this check
            if ((parms.TypeFlags.Flags1 == 0x03) && (parms.TypeFlags.Flags2 == 0x20000) &&
                (parms.TypeFlags.Flags4 == 0x0400) && (parms.TypeFlags.Flags3 != 0x07))
            {
                return new SR2PS2ARGBRAWImage(path, parms, segParms);
            }
            if (
                ((parms.TypeFlags.Flags1 == 0x01) || (parms.TypeFlags.Flags2 == 0x02))
                && (parms.BPPImage == 16)
                )
            {
                return new TRLPSPARGBRAWImage(path, parms);
            }
            // this is another nasty check which is necessary because the PS2 version of Defiance 
            // uses a nearly identical header
            if (
                ((parms.TypeFlags.Flags1 == 0x0C) || (parms.TypeFlags.Flags1 == 0x0200))
                && (parms.BPPImage == 8) && (parms.TypeFlags.Flags4 > 0x03FF)
                )
            {
                return new TRLPSPIndexedRAWImage(path, parms);
            }
            // and in fact the PS2 version of Defiance has its own detection problems
            if (
                ((parms.TypeFlags.Flags1 == 0x03) || (parms.TypeFlags.Flags1 == 0x00))
                && ((parms.TypeFlags.Flags2 == 0x10) || (parms.TypeFlags.Flags2 == 0x70))
                )
            {
                return new DefiancePS2ARGBRAWImage(path, parms, segParms);
            }
            if ((parms.TypeFlags.Flags1 == 0x0C) && (parms.TypeFlags.Flags2 == 0x0200))
            {
                return new DefiancePS2IndexedRAWImage(path, parms, segParms);
            }
            // as it turns out, the RAW type with the simplest format is also the hardest
            // to reliably ID
            if (
                ((parms.TypeFlags.Flags1 == 0x03) || (parms.TypeFlags.Flags1 == 0x00))
                && (parms.TypeFlags.Flags2 != 0x10)
                && (parms.TypeFlags.Flags2 != 0x70)
                )
            {
                return new SR2PCRAWImage(path, parms);
            }

            throw new RAWImageException("The file '" + path + "' has a RAW header, but does not appear to be in a supported format.");
        }

        protected static RAWHeaderSignature GetHeaderData(string path)
        {
            RAWHeaderSignature signature = RAWHeaderSignature.Unknown;
            uint header1, header2;

            FileStream iStream = new FileStream(path, FileMode.Open, FileAccess.Read);
            BinaryReader iReader = new BinaryReader(iStream);
            header1 = iReader.ReadUInt32();
            header2 = iReader.ReadUInt32();
            iReader.Close();
            iStream.Close();
            switch (header1)
            {
                case 0x52415721:
                    switch (header2)
                    {
                        case 0x54:
                            signature = RAWHeaderSignature.SR2PS2AirForgeDemo;
                            break;
                        default:
                            signature = RAWHeaderSignature.Standard;
                            break;
                    }
                    break;
                default:
                    signature = RAWHeaderSignature.Unknown;
                    break;
            }

            return signature;
        }

        protected virtual byte[] ReadBytes(long offset, int numBytes)
        {
            byte[] bytes = new byte[0];

            try
            {
                FileStream iStream = new FileStream(_FilePath, FileMode.Open, FileAccess.Read);
                BinaryReader iReader = new BinaryReader(iStream);
                iStream.Seek(offset, SeekOrigin.Begin);
                bytes = iReader.ReadBytes(numBytes);
                iReader.Close();
                iStream.Close();
            }
            catch (Exception ex)
            {
                throw new RAWImageException("An error occurred while loading the image data from the file.", ex);
            }

            return bytes;
        }

        public Bitmap ReadSingleARGBImageData()
        {
            byte[] rawData = ReadDataFromFile(_FilePath, _DataOffset, _RAWParameters.Width * _RAWParameters.Height * (BPPImage / 8));
            ImageData channelData = new ImageData(this);
            channelData.GetChannelsFromRawBytes(rawData, IgnoreAlpha);
            return channelData.CreateImage();
        }

        protected virtual Bitmap CreateBitmap(int width, int height, int stride, PixelFormat format, byte[] imageData)
        {
            try
            {
                Bitmap image = new Bitmap(width, height, format);
                Rectangle bounds = new Rectangle(0, 0, width, height);
                BitmapData data = image.LockBits(bounds, ImageLockMode.WriteOnly, format);
                Marshal.Copy(imageData, 0, data.Scan0, imageData.Length);
                //int offset = 0;
                //int bytesPerPixel = GetBytesPerPixelFromPixelFormat(format);
                //long ptr = data.Scan0.ToInt64();
                //for (int i = 0; i < height; i++)
                //{
                //    Marshal.Copy(imageData, offset, new IntPtr(ptr), width * bytesPerPixel);
                //    offset += width * bytesPerPixel;
                //    ptr += data.Stride;
                //}
                image.UnlockBits(data);
                return image;
            }
            catch (Exception ex)
            {
                throw new RAWImageException("Unable to create a bitmap from the image data.", ex);
            }

            //try
            //{
            //    unsafe
            //    {
            //        fixed (byte* imageBytesPtr = imageData)
            //        {
            //            IntPtr imageBytesPointer = new IntPtr((void*)imageBytesPtr);

            //            return new Bitmap(width, height, stride, format, imageBytesPointer);
            //        }
            //    }
            //}
            //catch (Exception ex)
            //{
            //    throw new RAWImageException("Unable to create a bitmap from the image data.", ex);
            //}
        }

        /// <summary>
        /// This will not handle formats with less than 1 byte per pixel correctly!
        /// </summary>
        /// <param name="format"></param>
        /// <returns></returns>
        public static int GetBytesPerPixelFromPixelFormat(PixelFormat format)
        {
            int result = 1;

            switch (format)
            {
                case PixelFormat.Format16bppArgb1555:
                    result = 2;
                    break;
                case PixelFormat.Format16bppGrayScale:
                    result = 2;
                    break;
                case PixelFormat.Format16bppRgb555:
                    result = 2;
                    break;
                case PixelFormat.Format16bppRgb565:
                    result = 2;
                    break;
                case PixelFormat.Format24bppRgb:
                    result = 3;
                    break;
                case PixelFormat.Format32bppArgb:
                    result = 4;
                    break;
                case PixelFormat.Format32bppPArgb:
                    result = 4;
                    break;
                case PixelFormat.Format32bppRgb:
                    result = 4;
                    break;
                case PixelFormat.Format48bppRgb:
                    result = 6;
                    break;
                case PixelFormat.Format64bppArgb:
                    result = 8;
                    break;
                case PixelFormat.Format64bppPArgb:
                    result = 8;
                    break;
            }

            return result;
        }

        #region File I/O

        protected byte[] ReadDataFromFile(string inFile, long offset, long numBytes)
        {
            FileStream fStream;
            BinaryReader bReader;

            byte[] readData = new byte[numBytes];

            for (int i = 0; i < numBytes; i++)
            {
                readData[i] = (byte)0;
            }

            if (File.Exists(inFile))
            {
                //try
                //{
                fStream = new FileStream(inFile, FileMode.Open, FileAccess.Read);
                bReader = new BinaryReader(fStream);
                fStream.Seek(offset, SeekOrigin.Begin);
                fStream.Read(readData, 0, (int)numBytes);

                bReader.Close();
                fStream.Close();
                return readData;
                //}
                //catch (EndOfStreamException ex)
                //{
                //    //MessageBox.Show("Reached end of file " + inFile + " before reading enough bytes", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                //}
                //catch (IOException ex)
                //{
                //    //MessageBox.Show("Error reading file: " + inFile, "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                //}
            }
            else
            {
                //MessageBox.Show("File not found: " + inFile, "File Not Found", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            return readData;
        }

        protected byte[] ReadDefiancePCDataFromFile(string inFile, long offset, long numBytes)
        {
            //reads the proprietary RLE-compressed data
            FileStream fStream;
            BinaryReader bReader;
            uint byteBuffer = 0;

            byte[] readData = new byte[numBytes];

            if (File.Exists(inFile))
            {
                try
                {
                    fStream = new FileStream(inFile, FileMode.Open, FileAccess.Read);
                    bReader = new BinaryReader(fStream);
                    fStream.Seek(offset, SeekOrigin.Begin);
                    int i = 0;
                    do
                    {
                        byteBuffer = bReader.ReadUInt32();
                        int runLength = 0;
                        uint byte1 = byteBuffer;
                        uint byte2 = byteBuffer;
                        uint byte3 = byteBuffer;
                        uint byte4 = byteBuffer;

                        byte1 <<= 24;
                        byte1 >>= 24;
                        byte2 <<= 16;
                        byte2 >>= 24;
                        byte3 <<= 8;
                        byte3 >>= 24;
                        byte4 >>= 24;

                        uint runCount1 = byte1;
                        uint runCount2 = byte2;
                        uint runCount3 = byte3;
                        uint runCount4 = byte4;

                        byte1 &= 0xFE;
                        byte2 &= 0xFE;
                        byte3 &= 0xFE;
                        byte4 &= 0xFE;

                        runCount1 &= 1;
                        runCount2 &= 1;
                        runCount3 &= 1;
                        runCount4 &= 1;

                        runCount2 <<= 1;
                        runCount3 <<= 2;
                        runCount4 <<= 3;

                        runLength = (int)(runCount1 | runCount2 | runCount3 | runCount4);
                        runLength++;

                        for (int j = 0; j < runLength; j++)
                        {
                            int jFour = j * 4;
                            readData[i + jFour] = (byte)byte4;
                            readData[i + jFour + 1] = (byte)byte3;
                            readData[i + jFour + 2] = (byte)byte2;
                            readData[i + jFour + 3] = (byte)byte1;
                        }
                        i += (4 * runLength);
                    } while (i < numBytes);
                    //fStream.Read(readData, 0, (int)numBytes);

                    bReader.Close();
                    fStream.Close();
                    return readData;
                }
                catch (EndOfStreamException ex)
                {
                    //MessageBox.Show("Reached end of file " + inFile + " before reading enough bytes", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (IOException ex)
                {
                    //MessageBox.Show("Error reading file: " + inFile, "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            else
            {
                //MessageBox.Show("File not found: " + inFile, "File Not Found", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            return readData;
        }
        #endregion

        #region Utility Functions

        protected Bitmap compositeImages(Bitmap mainImage, Bitmap addImage, int x, int y)
        {
            Bitmap tcImage;
            Color tColour;
            int tA, tR, tG, tB;
            int xRep, yRep;

            xRep = Math.Min(addImage.Width, mainImage.Width - x);
            yRep = Math.Min(addImage.Height, mainImage.Height - y);

            tcImage = new Bitmap(mainImage);

            for (int nY = 0; nY < yRep; nY++)
            {
                for (int nX = 0; nX < xRep; nX++)
                {
                    tColour = addImage.GetPixel(nX, nY);
                    tA = tColour.A;
                    tR = tColour.R;
                    tG = tColour.G;
                    tB = tColour.B;
                    tcImage.SetPixel(nX + x, nY + y, Color.FromArgb(tA, tR, tG, tB));
                }
            }

            return tcImage;
        }

        protected byte[] unSwizzleBytes(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int colCount = imgWidth / 16 * (imgBPP / 8);
            int rowCount = imgHeight / 8;

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int i = 0; i < rowCount; i++)
            {
                for (int j = 0; j < colCount; j++)
                {
                    for (int k = 0; k < 128; k++)
                    {
                        currentOutPos = (i * colCount * 128) + (j * 16) + (k % 16) + ((k / 16) * colCount * 16);
                        outData[currentOutPos] = inData[currentInPos];
                        currentInPos++;
                    }
                }
            }

            return outData;
        }

        protected byte[] unSwizzleBytes(byte[] inData, int imgWidth, int imgHeight, int imgBPP, int colX, int colY)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            int colCount = imgWidth / colX * (imgBPP / 8);
            int rowCount = imgHeight / colY;
            int colProduct = colX * colY;

            int currentInPos = 0;
            int currentOutPos = 0;

            for (int i = 0; i < rowCount; i++)
            {
                for (int j = 0; j < colCount; j++)
                {
                    for (int k = 0; k < colProduct; k++)
                    {
                        currentOutPos = (i * colCount * colProduct) + (j * colX) + (k % colX) + ((k / colX) * colCount * colX);
                        outData[currentOutPos] = inData[currentInPos];
                        currentInPos++;
                    }
                }
            }

            return outData;
        }

        protected byte[] FlipUInt16EndianOrder(byte[] inData)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            for (int i = 0; i < inData.GetUpperBound(0); i += 2)
            {
                outData[i] = inData[i + 1];
                outData[i + 1] = inData[i];
            }

            return outData;
        }

        protected byte[] FlipUInt32EndianOrder(byte[] inData)
        {
            byte[] outData;
            outData = new byte[inData.GetUpperBound(0) + 1];

            for (int i = 0; i < inData.GetUpperBound(0); i += 4)
            {
                outData[i] = inData[i + 3];
                outData[i + 1] = inData[i + 2];
                outData[i + 2] = inData[i + 1];
                outData[i + 3] = inData[i];
            }

            return outData;
        }

        protected byte[] unPS2SR2Swizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            //return inData;
            byte[] outData;
            outData = new byte[inData.Length];

            //int z = 0;
            //for (int y = 0; y < imgHeight; y++)
            //{
            //    for (int x = 1; x < imgWidth; x += 2)
            //    {
            //        outData[(y * imgWidth) + x] = inData[z];
            //        z++;
            //    }
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        outData[(y * imgWidth) + x] = inData[z];
            //        z++;
            //    }
            //}


            //this one is really close, but seems to need each meta-block of 4x4 pixels rotated too
            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        int inPos = (imgWidth * y) + x;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * y) + x + 1;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * (y + 1)) + x;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;

            //        inPos = (imgWidth * (y + 1)) + x + 1;
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //}

            int[] pixelIndex02 = new int[]
            {
                0, 1, 4, 5, 8, 9, 12, 13,       0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15,     2, 3, 6, 7, 10, 11, 14, 15,

                8, 9, 12, 13, 0, 1, 4, 5,       8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7,     10, 11, 14, 15, 2, 3, 6, 7 
            };

            int[] pixelIndex13 = new int[]
            {
                8, 9, 12, 13, 0, 1, 4, 5,       8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7,     10, 11, 14, 15, 2, 3, 6, 7, 

                0, 1, 4, 5, 8, 9, 12, 13,       0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15,     2, 3, 6, 7, 10, 11, 14, 15
            };

            //int[] subIndex = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,          32, 32, 32, 32, 32, 32, 32, 32,
            //    0, 0, 0, 0, 0, 0, 0, 0,          32, 32, 32, 32, 32, 32, 32, 32,

            //    16, 16, 16, 16, 16, 16, 16, 16,      48, 48, 48, 48, 48, 48, 48, 48,
            //    16, 16, 16, 16, 16, 16, 16, 16,      48, 48, 48, 48, 48, 48, 48, 48
            //};

            int[] subIndex = new int[]
            {
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,

                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8,
                0, 0, 0, 0, 0, 0, 0, 0,          8, 8, 8, 8, 8, 8, 8, 8
            };

            //int[] xIndex = new int[]
            //{
            //    0, 1, 
            //    0, 1
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0,
            //    1, 1
            //};

            //int[] xIndex = new int[]
            //{
            //    1, 0, 
            //    1, 0
            //};

            //int[] yIndex = new int[]
            //{
            //    1, 1,
            //    0, 0
            //};

            //int[] xIndex = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,

            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11,
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,              0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1,              1, 1, 1, 1, 1, 1, 1, 1, 
            //    2, 2, 2, 2, 2, 2, 2, 2,              2, 2, 2, 2, 2, 2, 2, 2, 
            //    3, 3, 3, 3, 3, 3, 3, 3,              3, 3, 3, 3, 3, 3, 3, 3
            //};

            //int[] xIndexA = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15,
            //    0, 1, 2, 3, 4, 5, 6, 7,             8, 9, 10, 11, 12, 13, 14, 15
            //};

            //int[] xIndexA = new int[]
            //{
            //    0, 1, 2, 3, 4, 5, 6, 7,
            //    0, 1, 2, 3, 4, 5, 6, 7
            //};

            //int[] xIndexB = new int[]
            //{
            //    4, 5, 6, 7, 0, 1, 2, 3,
            //    4, 5, 6, 7, 0, 1, 2, 3
            //};

            //int[] yIndex2 = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1
            //};

            //int[] xIndexB = new int[]
            //{
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11,
            //    4, 5, 6, 7, 0, 1, 2, 3,             12, 13, 14, 15, 8, 9, 10, 11
            //};

            //int[] yIndex2 = new int[]
            //{
            //    0, 0, 0, 0, 0, 0, 0, 0,              0, 0, 0, 0, 0, 0, 0, 0,
            //    1, 1, 1, 1, 1, 1, 1, 1,              1, 1, 1, 1, 1, 1, 1, 1, 
            //};

            //int[] blockIndex = new int[]
            //{
            //    0, 1, 4, 5,         16, 17, 20, 21,
            //    2, 3, 6, 7,         18, 19, 22, 23,
            //    8, 9, 12, 13,       24, 25, 28, 29,
            //    10, 11, 14, 15,     26, 27, 30, 31
            //};

            //int[] blockIndex16 = new int[]
            //{
            //    0, 1, 4, 5,
            //    2, 3, 6, 7,
            //    8, 9, 12, 13,
            //    10, 11, 14, 15
            //};

            //int[] blockIndex4 = new int[]
            //{
            //    0, 1,
            //    2, 3
            //};

            //int[] xIndex = new int[]
            //{
            //    0, 1, 2, 3, 
            //    0, 1, 2, 3,
            //    0, 1, 2, 3,
            //    0, 1, 2, 3
            //};

            //int[] yIndex = new int[]
            //{
            //    0, 0, 0, 0,
            //    1, 1, 1, 1,
            //    2, 2, 2, 2,
            //    3, 3, 3, 3
            //};

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 2)
            //    {
            //        for (int i = 0; i < 4; i++)
            //        {
            //            int inPos = (imgWidth * (y + yIndex[i])) + x + xIndex[i];
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //int[] bitMap = new int[]
            //{
            //    6, 7, 8, 9,
            //    5, 0, 1, 10,
            //    4, 3, 2, 11,
            //    15, 14, 13, 12
            //};

            int[] bitMap0 = new int[]
            {
                0, 1, 4, 5, 8, 9, 12, 13,
                2, 3, 6, 7, 10, 11, 14, 15
            };

            int[] bitMap1 = new int[]
            {
                8, 9, 12, 13, 0, 1, 4, 5,
                10, 11, 14, 15, 2, 3, 6, 7
            };

            //int inPos = 0;
            int globalX = 0;
            int globalY = 0;
            int chunkX = 8;
            int chunkY = 2;
            int chunkSize = chunkX * chunkY;
            int[] bitMap = bitMap0;
            int whichBitMap = 0;
            for (int inPos = 0; inPos < inData.Length; inPos += chunkSize)
            {
                //byte[] localBytes = new byte[chunkSize];
                //for (int i = 0; i < localBytes.Length; i++)
                //{
                //    localBytes[i] = inData[inPos + i];
                //}
                int iterator = 0;
                for (int y = 0; y < chunkY; y++)
                {
                    for (int x = 0; x < chunkX; x++)
                    {
                        int mappedPosition = bitMap[(y * chunkX) + x];
                        int localX = globalX + (mappedPosition % chunkX);
                        int localY = globalY + (mappedPosition / chunkX);
                        int outPos = (localY * imgWidth) + localX;
                        outData[outPos] = inData[inPos + iterator];
                        iterator++;
                    }
                }
                globalX += chunkX;
                if (globalX >= imgWidth)
                {
                    whichBitMap++;
                    if (whichBitMap > 3)
                    {
                        whichBitMap = 0;
                    }
                    switch (whichBitMap)
                    {
                        case 0:
                            bitMap = bitMap1;
                            break;
                        case 1:
                            bitMap = bitMap1;
                            break;
                        case 2:
                            bitMap = bitMap1;
                            break;
                        case 3:
                            bitMap = bitMap1;
                            break;
                    }
                    globalX = 0;
                    globalY += chunkY;
                }
            }
            //int inPosBase = 0;
            //for (int y = 0; y < imgHeight; y += 8)
            //{
            //    inPosBase = y * imgWidth;
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int j = 0; j < 64; j++)
            //        {
            //            //int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            int inPos = inPosBase + x + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            outData[outPos] = inData[inPos];
            //            //outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //        //inPosBase += 64;
            //    }
            //    inPosBase = (y + 4) * imgWidth;
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int j = 0; j < 64; j++)
            //        {
            //            //int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            int inPos = inPosBase + x + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //            outData[outPos] = inData[inPos];
            //            //outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //        //inPosBase += 64;
            //    } 
            //    //for (int i = 0; i < inData.Length; i += 128)
            //    //{
            //    //    inPosBase = i;
            //    //    for (int j = 0; j < 64; j++)
            //    //    {
            //    //        int inPos = inPosBase + pixelIndex02[j] + subIndex[j];
            //    //        outData[outPos] = inData[inPos];
            //    //        //outData[inPos] = inData[outPos];
            //    //        outPos++;
            //    //    }
            //    //    inPosBase += 64;
            //    //    for (int j = 0; j < 64; j++)
            //    //    {
            //    //        int inPos = inPosBase + pixelIndex13[j] + subIndex[j];
            //    //        outData[outPos] = inData[inPos];
            //    //        //outData[inPos] = inData[outPos];
            //    //        outPos++;
            //    //    }
            //    //}
            //}
            //for (int i = 0; i < inData.Length; i += 128)
            //{
            //    inPosBase = i;
            //    for (int j = 0; j < 64; j++)
            //    {
            //        int inPos = inPosBase + pixelIndex02[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //    inPosBase += 64;
            //    for (int j = 0; j < 64; j++)
            //    {
            //        int inPos = inPosBase + pixelIndex13[j] + subIndex[j] + (imgWidth * yIndex[j]);
            //        //outData[outPos] = inData[inPos];
            //        outData[inPos] = inData[outPos];
            //        outPos++;
            //    }
            //}
            //int inPos = 0;
            //for (int y = 0; y < imgHeight; y += 4)
            //{
            //    for (int i = 0; i < (imgWidth * 2); i += 16)
            //    {
            //        int startX = i / 2;
            //        for (int k = 0; k < 4; k++)
            //        {
            //            int currentX = startX + (k * 2);
            //            for (int j = 0; j < 4; j++)
            //            {
            //                int outPos = (y * imgWidth) + (yIndex[j] * imgWidth) + currentX + xIndex[j];
            //                outData[outPos] = inData[inPos];
            //                inPos++;
            //            }
            //        }
            //    }
            //    for (int i = 0; i < (imgWidth * 2); i += 16)
            //    {
            //        int startX = i / 2;
            //        for (int k = 0; k < 4; k++)
            //        {
            //            int currentX = startX + (k * 2);
            //            for (int j = 0; j < 4; j++)
            //            {
            //                int outPos = ((y + 2) * imgWidth) + (yIndex[j] * imgWidth) + currentX + xIndex[j];
            //                outData[outPos] = inData[inPos];
            //                inPos++;
            //            }
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 4)
            //    {
            //        for (int j = 0; j < 4; j++)
            //        {
            //            for (int i = 0; i < 4; i++)
            //            {
            //                int inPos = (imgWidth * (y + yIndex[i])) + x + xIndex[i];
            //                inPos += (xIndex[j] * 2);
            //                //inPos += (yIndex[j] * 2);
            //                outData[inPos] = inData[outPos];
            //                outPos++;
            //            }
            //        }
            //    }
            //}

            //int numBlocks = (inData.Length / (128 * 64));
            //for (int blockNum = 0; blockNum < numBlocks; blockNum++)
            //{
            //    for (int colNum = 0; colNum < 8; colNum++)
            //    {
            //        for (int z = 0; z < (2 * 16); z++)
            //        {
            //            int xFactor = 0;
            //            if (colNum % 2 == 1)
            //            {
            //                xFactor = xIndexB[z];
            //            }
            //            else
            //            {
            //                xFactor = xIndexA[z];
            //            }
            //            xFactor += (yIndex[z] * imgWidth);
            //            int inPos = 0;
            //            inPos = (blockNum * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            if (numBlocks < 5)
            //            {
            //                inPos = (blockIndex4[blockNum] * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            }
            //            else
            //            {
            //                inPos = (blockIndex16[blockNum] * (128 * 64)) + (colNum * (16 * 2)) + xFactor;
            //            }
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //int arraySwitch = 0;

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    for (int x = 0; x < imgWidth; x += 16)
            //    {
            //        for (int i = 0; i < (2 * 16); i++)
            //        {
            //            int xFactor = 0;
            //            switch (arraySwitch)
            //            {
            //                case 0:
            //                    xFactor = xIndexA[i];
            //                    break;
            //                case 1:
            //                    xFactor = xIndexB[i];
            //                    break;
            //                case 2:
            //                    xFactor = xIndexB[i];
            //                    break;
            //                case 3:
            //                    xFactor = xIndexA[i];
            //                    break;
            //                default:
            //                    throw new Exception("Error");
            //                    break;
            //            }
            //            arraySwitch++;
            //            if (arraySwitch > 3)
            //            {
            //                arraySwitch = 0;
            //            }
            //            int inPos = (imgWidth * (y + yIndex[i])) + x + xFactor;
            //            outData[inPos] = inData[outPos];
            //            outPos++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight / 4; y += 4)
            //{
            //    for (int x = 0; x < imgWidth / 16; x += 16)
            //    {
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = pixelIndex02[i] + subIndex[i];
            //            inPos += (y * imgWidth) + x;
            //            outData[inPos] = inData[outPos];
            //        }
            //    }
            //}

            //int byteNum = 0;

            //for (int blockY = 0; blockY < imgHeight / 16; blockY++)
            //{
            //    for (int blockX = 0; blockX < imgWidth / 16; blockX++)
            //    {
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex02[i] + subIndex[i];
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex13[i] + subIndex[i] + 64;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex02[i] + subIndex[i] + 128;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //        for (int i = 0; i < 64; i++)
            //        {
            //            int inPos = (blockY * imgWidth * 16) + (blockX * 16) + pixelIndex13[i] + subIndex[i] + 192;
            //            //outData[inPos] = inData[byteNum];
            //            outData[byteNum] = inData[inPos];
            //            byteNum++;
            //        }
            //    }
            //}

            //for (int block = 0; block < inData.Length / 256; block++)
            //{
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex02[i] + subIndex[i]];
            //        outData[(block * 256) + pixelIndex02[i] + subIndex[i]] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex13[i] + subIndex[i] + 64];
            //        outData[(block * 256) + pixelIndex13[i] + subIndex[i] + 64] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex02[i] + subIndex[i] + 128];
            //        outData[(block * 256) + pixelIndex02[i] + subIndex[i] + 128] = inData[byteNum];
            //        byteNum++;
            //    }
            //    for (int i = 0; i < 64; i++)
            //    {
            //        //outData[byteNum] = inData[(block * 256) + pixelIndex13[i] + subIndex[i] + 192];
            //        outData[(block * 256) + pixelIndex13[i] + subIndex[i] + 192] = inData[byteNum];
            //        byteNum++;
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 32;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = div; xx < imgWidth; xx += numDivs)
            //        {
            //            outData[(y * imgWidth) + x] = inData[(y * imgWidth + xx)];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 4;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = div; xx < imgWidth; xx += numDivs)
            //        {
            //            outData[(y * imgWidth) + xx] = inData[(y * imgWidth + x)];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 64;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = 0; xx < divWidth; xx++)
            //        {
            //            int outPixelX = (div * divWidth) + (divWidth - 1 - xx);
            //            int inPixelX = xx + (div * divWidth);
            //            //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            x++;
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    int x = 0;
            //    int numDivs = 4;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        for (int xx = 0; xx < divWidth; xx++)
            //        {
            //            int outPixelX = (div * divWidth) + (divWidth - 1 - xx);
            //            int inPixelX = xx + (div * divWidth);
            //            //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //            x++;
            //        }
            //    }
            //}

            //int outPos = 0;

            //for (int y = 0; y < imgHeight; y += 2)
            //{
            //    int divCount = 1;
            //    int divWidth = imgWidth / divCount;
            //    for (int div = 0; div < divCount; div++)
            //    {
            //        //for (int x = 0; x < (divWidth / 2); x++)
            //        //{
            //        //    outData[outPos] = inData[(y * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}
            //        //for (int x = 0; x < (divWidth / 2); x++)
            //        //{
            //        //    outData[outPos] = inData[((y + 1) * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}

            //        //for (int x = (divWidth / 2); x < divWidth; x++)
            //        //{
            //        //    outData[outPos] = inData[(y * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}
            //        //for (int x = (divWidth / 2); x < divWidth; x++)
            //        //{
            //        //    outData[outPos] = inData[((y + 1) * imgWidth) + (div * divWidth) + x];
            //        //    outPos++;
            //        //}

            //        for (int x = 0; x < (divWidth / 2); x++)
            //        {
            //            outData[(y * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }
            //        for (int x = 0; x < (divWidth / 2); x++)
            //        {
            //            outData[((y + 1) * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }

            //        for (int x = (divWidth / 2); x < divWidth; x++)
            //        {
            //            outData[(y * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }
            //        for (int x = (divWidth / 2); x < divWidth; x++)
            //        {
            //            outData[((y + 1) * imgWidth) + (div * divWidth) + x] = inData[outPos];
            //            outPos++;
            //        }




            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    //int x = 0;
            //    int numDivs = 8;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        int numSubDivs = 8;
            //        int subDivWidth = divWidth / numSubDivs;
            //        for (int subDiv = 0; subDiv < numSubDivs; subDiv++)
            //        {
            //            int x = 0;
            //            for (int xx = 0; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //                x++;
            //            }
            //            for (int xx = 1; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //                x++;
            //            }

            //            //for (int xx = 0; xx < subDivWidth; xx++)
            //            //{
            //            //    int outPixelX = subDivWidth - 1 - xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            //    outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            //    x++;
            //            //}
            //        }
            //    }
            //}

            //for (int y = 0; y < imgHeight; y++)
            //{
            //    //int x = 0;
            //    int numDivs = 16;
            //    int divWidth = imgWidth / numDivs;
            //    for (int div = 0; div < numDivs; div++)
            //    {
            //        int numSubDivs = 2;
            //        int subDivWidth = divWidth / numSubDivs;
            //        for (int subDiv = 0; subDiv < numSubDivs; subDiv++)
            //        {
            //            int x = 0;
            //            for (int xx = 0; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //                x++;
            //            }
            //            for (int xx = 1; xx < subDivWidth; xx += 2)
            //            {
            //                int outPixelX = x + (div * divWidth) + (subDiv * subDivWidth);
            //                int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //                //int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //                //finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //                outData[(y * imgWidth) + inPixelX] = inData[(y * imgWidth) + outPixelX];
            //                x++;
            //            }

            //            //for (int xx = 0; xx < subDivWidth; xx++)
            //            //{
            //            //    int outPixelX = subDivWidth - 1 - xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    int outPixelX = (div * divWidth) + (subDiv * subDivWidth) + (divWidth - 1 - xx);
            //            //    int inPixelX = xx + (div * divWidth) + (subDiv * subDivWidth);
            //            //    finalImage.SetPixel(outPixelX, y, palette.GetPixel(tempImage.GetPixel(inPixelX, y).G, 0));
            //            //    outData[(y * imgWidth) + outPixelX] = inData[(y * imgWidth) + inPixelX];
            //            //    x++;
            //            //}
            //        }
            //    }
            //}

            //int currentInPos = 0;
            //int currentOutPos = 0;

            //int colCount = imgWidth / 128;
            //int rowCount = imgHeight / 64;

            //for (int i = 0; i < rowCount; i++)
            //{
            //    for (int j = 0; j < colCount; j++)
            //    {
            //        for (int k = 0; k < 128; k++)
            //        {
            //            currentOutPos = (i * colCount * 128) + (j * 16) + (k % 16) + ((k / 16) * colCount * 16);
            //            outData[currentOutPos] = inData[currentInPos];
            //            currentInPos++;
            //        }
            //    }
            //}

            //// 2/2 is close
            //int xFactor = 16;
            //int yFactor = 16;
            //int blockSize = xFactor * yFactor;

            //int colCount = imgWidth / xFactor * (imgBPP / 8);
            //int rowCount = imgHeight / yFactor;

            //int[][] pixelIndex = BuildPSMT8PixelIndex();


            //for (int i = 0; i < rowCount; i++)
            //{
            //    for (int j = 0; j < colCount; j++)
            //    {
            //        for (int k = 0; k < blockSize; k++)
            //        {
            //            //currentOutPos = (i * colCount * blockSize) + (j * xFactor) + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            //currentInPos = (i * colCount * blockSize) + (j * xFactor) + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            //              ^^^for each entire row        ^h to block     ^x in block      ^y in block    
            //            //currentOutPos = (i * colCount * blockSize) + (j * xFactor) + (k % 64) + ((k / 64) * colCount * 16);
            //            int currentBasePos = (i * colCount * blockSize) + (j * xFactor);
            //            currentOutPos = currentBasePos + (k % xFactor) + ((k / xFactor) * colCount * xFactor);
            //            currentInPos = currentBasePos + (pixelIndex[k % 4][k % 64]) + ((pixelIndex[k / xFactor][k] / xFactor) * colCount * xFactor);//pixelIndex[k / 16][k];
            //            outData[currentOutPos] = inData[currentInPos];
            //            currentOutPos++;
            //        }
            //    }
            //}

            return outData;
        }

        protected byte[] unPS2DefianceSwizzle(byte[] inData, int imgWidth, int imgHeight, int imgBPP)
        {
            //return inData;
            byte[] outData;
            //float fudgePercent = 1.1f;
            outData = new byte[inData.Length];
            //outData = new byte[(int)(inData.Length * fudgePercent)];

            Hashtable map1 = new Hashtable();
            Hashtable map2 = new Hashtable();

            map1.Add(0, new Point(0, 0));
            map1.Add(1, new Point(4, 1));
            map1.Add(2, new Point(8, 0));
            map1.Add(3, new Point(12, 1));
            map1.Add(4, new Point(1, 0));
            map1.Add(5, new Point(5, 1));
            map1.Add(6, new Point(9, 0));
            map1.Add(7, new Point(13, 1));
            map1.Add(8, new Point(2, 0));
            map1.Add(9, new Point(6, 1));
            map1.Add(10, new Point(10, 0));
            map1.Add(11, new Point(14, 1));
            map1.Add(12, new Point(3, 0));
            map1.Add(13, new Point(7, 1));
            map1.Add(14, new Point(11, 0));
            map1.Add(15, new Point(15, 1));
            map1.Add(16, new Point(4, 0));
            map1.Add(17, new Point(0, 1));
            map1.Add(18, new Point(12, 0));
            map1.Add(19, new Point(8, 1));
            map1.Add(20, new Point(5, 0));
            map1.Add(21, new Point(1, 1));
            map1.Add(22, new Point(13, 0));
            map1.Add(23, new Point(9, 1));
            map1.Add(24, new Point(6, 0));
            map1.Add(25, new Point(2, 1));
            map1.Add(26, new Point(14, 0));
            map1.Add(27, new Point(10, 1));
            map1.Add(28, new Point(7, 0));
            map1.Add(29, new Point(3, 1));
            map1.Add(30, new Point(15, 0));
            map1.Add(31, new Point(11, 1));

            map2.Add(0, new Point(4, 0));
            map2.Add(1, new Point(0, 1));
            map2.Add(2, new Point(12, 0));
            map2.Add(3, new Point(8, 1));
            map2.Add(4, new Point(5, 0));
            map2.Add(5, new Point(1, 1));
            map2.Add(6, new Point(13, 0));
            map2.Add(7, new Point(9, 1));
            map2.Add(8, new Point(6, 0));
            map2.Add(9, new Point(2, 1));
            map2.Add(10, new Point(14, 0));
            map2.Add(11, new Point(10, 1));
            map2.Add(12, new Point(7, 0));
            map2.Add(13, new Point(3, 1));
            map2.Add(14, new Point(15, 0));
            map2.Add(15, new Point(11, 1));
            map2.Add(16, new Point(0, 0));
            map2.Add(17, new Point(4, 1));
            map2.Add(18, new Point(8, 0));
            map2.Add(19, new Point(12, 1));
            map2.Add(20, new Point(1, 0));
            map2.Add(21, new Point(5, 1));
            map2.Add(22, new Point(9, 0));
            map2.Add(23, new Point(13, 1));
            map2.Add(24, new Point(2, 0));
            map2.Add(25, new Point(6, 1));
            map2.Add(26, new Point(10, 0));
            map2.Add(27, new Point(14, 1));
            map2.Add(28, new Point(3, 0));
            map2.Add(29, new Point(7, 1));
            map2.Add(30, new Point(11, 0));
            map2.Add(31, new Point(15, 1));

            int[] map = new int[]
            {        
                           
                0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15,
                16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31
            };

            int inPos = 0;

            for (int y = 0; y < imgHeight; y += 8)
            {
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y;
                        int xPos = x;

                        Point mapPoint = (Point)map1[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        //outData[(yPos * imgWidth) + xPos] = inData[inPos];

                        int outPos = (yPos * imgWidth) + xPos;
                        if ((outPos < outData.Length) && (inPos < inData.Length))
                        {
                            outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        }

                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 2;
                        int xPos = x;

                        Point mapPoint = (Point)map1[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        //outData[(yPos * imgWidth) + xPos] = inData[inPos];

                        int outPos = (yPos * imgWidth) + xPos;
                        if ((outPos < outData.Length) && (inPos < inData.Length))
                        {
                            outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        }

                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 4;
                        int xPos = x;

                        Point mapPoint = (Point)map2[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        //outData[(yPos * imgWidth) + xPos] = inData[inPos];

                        int outPos = (yPos * imgWidth) + xPos;
                        if ((outPos < outData.Length) && (inPos < inData.Length))
                        {
                            outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        }

                        inPos++;
                    }
                }
                for (int x = 0; x < imgWidth; x += 16)
                {
                    for (int i = 0; i < 32; i++)
                    {
                        int yPos = y + 6;
                        int xPos = x;

                        Point mapPoint = (Point)map2[map[i]];

                        xPos += mapPoint.X;
                        yPos += mapPoint.Y;

                        //outData[(yPos * imgWidth) + xPos] = inData[inPos];

                        int outPos = (yPos * imgWidth) + xPos;
                        if ((outPos < outData.Length) && (inPos < inData.Length))
                        {
                            outData[(yPos * imgWidth) + xPos] = inData[inPos];
                        }

                        inPos++;
                    }
                }
            }

            //flip rows
            for (int y = 0; y < imgHeight; y += 4)
            {
                byte[] rowBuffer = new byte[imgWidth];
                //copy second row to buffer
                for (int i = 0; i < imgWidth; i++)
                {
                    rowBuffer[i] = outData[((y + 1) * imgWidth) + i];
                }
                //copy third row to second
                for (int i = 0; i < imgWidth; i++)
                {
                    outData[((y + 1) * imgWidth) + i] = outData[((y + 2) * imgWidth) + i];
                }
                //copy buffer to third row
                for (int i = 0; i < imgWidth; i++)
                {
                    outData[((y + 2) * imgWidth) + i] = rowBuffer[i];
                }
            }

            return outData;
        }


        protected int[] BuildPSMT8BlockIndex()
        {
            int[] blockIndex = new int[] 
            {
                0,  1,  4,  5,  16, 17, 20, 21,
                2,  3,  6,  7,  18, 19, 22, 23,
                8,  9,  12, 13, 24, 25, 28, 29,
                10, 11, 14, 15, 26, 27, 30, 31
            };
            return blockIndex;
        }

        protected int[] BuildPSMT8ColumnIndex()
        {
            int[] columnIndex = new int[]
            {
                0,
                1,
                2,
                3
            };
            return columnIndex;
        }

        protected int[] BuildPSMT8ByteIndex()
        {
            int[] byteIndex = new int[]
            {
                0,  0,  0,  0,  0,  0,  0,  0,      2,  2,  2,  2,  2,  2,  2,  2,
                0,  0,  0,  0,  0,  0,  0,  0,      2,  2,  2,  2,  2,  2,  2,  2,

                1,  1,  1,  1,  1,  1,  1,  1,      3,  3,  3,  3,  3,  3,  3,  3,
                1,  1,  1,  1,  1,  1,  1,  1,      3,  3,  3,  3,  3,  3,  3,  3
            };

            return byteIndex;
        }

        protected int[][] BuildPSMT8PixelIndex()
        {
            int[] pixelIndex02 = new int[]
            {
                0,  1,  4,  5,  8,  9,  12, 13,     0,  1,  4,  5,  8,  9,  12, 13,
                2,  3,  6,  7,  10, 11, 14, 15,     2,  3,  6,  7,  10, 11, 14, 15,

                8,  9,  12, 13, 0,  1,  4,  5,      8,  9,  12, 13, 0,  1,  4,  5, 
                10, 11, 14, 15, 2,  3,  6,  7,      10, 11, 14, 15, 2,  3,  6,  7
            };

            int[] pixelIndex13 = new int[]
            {
                8,  9,  12, 13, 0,  1,  4,  5,      8,  9,  12, 13, 0,  1,  4,  5, 
                10, 11, 14, 15, 2,  3,  6,  7,      10, 11, 14, 15, 2,  3,  6,  7,

                0,  1,  4,  5,  8,  9,  12, 13,     0,  1,  4,  5,  8,  9,  12, 13,
                2,  3,  6,  7,  10, 11, 14, 15,     2,  3,  6,  7,  10, 11, 14, 15
            };

            int[][] pixelIndex = new int[4][];
            pixelIndex[0] = pixelIndex02;
            pixelIndex[1] = pixelIndex13;
            pixelIndex[2] = pixelIndex02;
            pixelIndex[3] = pixelIndex13;

            //int[] byteIndex = BuildPSMT8ByteIndex();

            //int[] tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex02[j] + (16 * byteIndex[j]);
            //}

            //pixelIndex[0] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex13[j] + (16 * byteIndex[j]) + 64;
            //}

            //pixelIndex[1] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex02[j] + (16 * byteIndex[j]) + 128;
            //}

            //pixelIndex[2] = tempIndex;

            //tempIndex = new int[64];

            //for (int j = 0; j < 64; j++)
            //{
            //    tempIndex[j] = pixelIndex13[j] + (16 * byteIndex[j]) + 192;
            //}

            //pixelIndex[3] = tempIndex;

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i] = pixelIndex02[i] + (16 * byteIndex[i]);
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 64] = pixelIndex13[i] + (16 * byteIndex[i]) + 64;
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 128] = pixelIndex02[i] + (16 * byteIndex[i]) + 128;
            ////}

            ////for (int i = 0; i < 64; i++)
            ////{
            ////    pixelIndex[i + 192] = pixelIndex13[i] + (16 * byteIndex[i]) + 192;
            ////}

            return pixelIndex;
        }

        protected int[] BuildIDTex8ClutIndex()
        {
            int[] clutIndex = new int[256];
            int currentIndex = 0;
            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 16 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 8 + j] = currentIndex;
                    currentIndex++;
                }
                for (int j = 0; j < 8; j++)
                {
                    clutIndex[(i * 32) + 24 + j] = currentIndex;
                    currentIndex++;
                }
            }
            return clutIndex;
        }


        protected uint ReadUInt32FromFile(string filePath, long offset)
        {
            uint readUInt;
            FileStream iStream;
            BinaryReader iReader;

            iStream = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            iReader = new BinaryReader(iStream);
            iStream.Seek(offset, SeekOrigin.Begin);
            readUInt = iReader.ReadUInt32();
            iReader.Close();
            iStream.Close();
            return readUInt;
        }

        protected uint ReadMask(BinaryReader reader)
        {
            uint byte1 = reader.ReadByte();
            uint byte2 = reader.ReadByte();
            uint byte3 = reader.ReadByte();
            uint byte4 = reader.ReadByte();

            byte1 <<= 24;
            byte2 <<= 16;
            byte3 <<= 8;

            return byte1 | byte2 | byte3 | byte4;
        }

        //reverses a mask uint because the PC version of Defiance has backwards masks
        protected uint TransformMask(uint inMask)
        {
            uint byte1 = inMask & 0xFF;
            uint byte2 = inMask & 0xFF00;
            uint byte3 = inMask & 0xFF0000;
            uint byte4 = inMask & 0xFF000000;

            byte1 <<= 24;
            byte2 <<= 8;
            byte3 >>= 8;
            byte4 >>= 24;

            return byte1 | byte2 | byte3 | byte4;
        }

        protected int getBitPositionInDWord(uint inDWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 32; i++)
            {
                bitMask = (uint)(1 << (31 - i));
                masked = (inDWord & bitMask) >> (31 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

        protected int getBitPositionInWord(uint inWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 16; i++)
            {
                bitMask = (ushort)(1 << (15 - i));
                masked = (inWord & bitMask) >> (15 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

        #endregion

    }
}
