// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using BLD = BenLincoln.Data;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class SR2PCRAWImage : RAWImage
    {
        public SR2PCRAWImage(string path, RAWParameters parms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _TypeDescription = "Soul Reaver 2 (PC) RAW Bitmap";
            _FormatDescription = "Uncompressed ARGB Image";

            _DataOffset = 128;
            if (parms.BPPImage == 32)
            {
                _BPPAlpha = (int)parms.BPP1;
                _BPPRed = (int)parms.BPP4;
                _BPPGreen = (int)parms.BPP3;
                _BPPBlue = (int)parms.BPP2;

                _MaskAlpha = BLD.EndianConverter.ReverseUInt(parms.Mask1);
                _MaskRed = BLD.EndianConverter.ReverseUInt(parms.Mask4);
                _MaskGreen = BLD.EndianConverter.ReverseUInt(parms.Mask3);
                _MaskBlue = BLD.EndianConverter.ReverseUInt(parms.Mask2);
            }
            else
            {
                _BPPAlpha = (int)parms.BPP1;
                _BPPRed = (int)parms.BPP2;
                _BPPGreen = (int)parms.BPP3;
                _BPPBlue = (int)parms.BPP4;

                _MaskAlpha = parms.Mask1;
                _MaskRed = parms.Mask2;
                _MaskGreen = parms.Mask3;
                _MaskBlue = parms.Mask4;
            }
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            byte[] imageBytes = ReadBytes(DataOffset, (int)DataLength);

            PixelFormat format = PixelFormat.Format32bppArgb;
            switch (BPPImage)
            {
                case 32:
                    format = PixelFormat.Format32bppArgb;
                    break;
                case 16:
                    format = PixelFormat.Format16bppArgb1555;
                    break;
                default:
                    throw new NotImplementedException("This type of RAW file (" +
                        TypeDescription + ") does not support this bit depth (" +
                        BPPImage.ToString() + ")");
                    break;
            }
            return CreateBitmap(Width, Height, _RAWParameters.Stride, format, imageBytes);
        }
    }
}
