// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class SR2PS2ARGBRAWImage : SegmentedRAWImage
    {
        public SR2PS2ARGBRAWImage(string path, RAWParameters parms, SegmentedRAWParameters segParms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _SegmentedRAWParameters = segParms;
            _TypeDescription = "Soul Reaver 2 (Playstation 2) RAW Bitmap";
            _FormatDescription = "Uncompressed Tiled RGB Image";

            _DataOffset = 144;
            BPPAlpha = (int)_RAWParameters.BPP1;
            BPPRed = (int)_RAWParameters.BPP2;
            BPPGreen = (int)_RAWParameters.BPP3;
            BPPBlue = (int)_RAWParameters.BPP4;
            MaskAlpha = _RAWParameters.Mask1;
            MaskRed = _RAWParameters.Mask2;
            MaskGreen = _RAWParameters.Mask3;
            MaskBlue = _RAWParameters.Mask4;
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            //throw new Exception("The method or operation is not implemented.");

            Bitmap tempImage = new Bitmap(_RAWParameters.Width, _RAWParameters.Height);

            int subHorizontalCount = (ushort)(_RAWParameters.Width / 256);
            if ((subHorizontalCount * 256) < _RAWParameters.Width)
            {
                subHorizontalCount++;
            }
            //int subVerticalCount = (ushort)(mImageHeight / 256);
            //if ((subVerticalCount * 256) < mImageHeight)
            //{
            //    subVerticalCount++;
            //}
            //int subImageCount = subHorizontalCount * subVerticalCount;


            int[] subImageWidths = new int[32];
            int[] subImageHeights = new int[32];
            int imageCounter = 0;
            long positionCounter = 128;
            int vOffset = 0;

            int rowCount = 0;
            do
            {
                int hOffset = 0;
                for (int j = 0; j < subHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, _RAWParameters.Width - hOffset);
                    //get current data length
                    long currentDataLength = ReadUInt32FromFile(_FilePath, positionCounter) * 16;

                    subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter] * (BPPImage / 8)));

                    long subImageOffset = positionCounter + 16;

                    positionCounter += 16 + currentDataLength;

                    byte[] rawData = ReadDataFromFile(_FilePath, subImageOffset, currentDataLength);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData, IgnoreAlpha);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                //vOffset += subImageHeights[imageCounter - 1];
                vOffset += subImageHeights[(rowCount * subHorizontalCount)];
                rowCount++;
            } while (vOffset < _RAWParameters.Height);

            return tempImage;
        }
    }
}
