// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class SR2PS2AirForgeIndexedRAWImage : RAWImage
    {
        public SR2PS2AirForgeIndexedRAWImage(string path, RAWParameters parms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _TypeDescription = "Soul Reaver 2 Air Forge Demo (Playstation 2) RAW Bitmap";
            _FormatDescription = "Uncompressed Indexed Image";
            IgnoreAlpha = true;

            _DataOffset = 84;
            // why is there an if statement here?
            if (BPPImage == 32)
            {
                BPPAlpha = (int)_RAWParameters.BPP1;
                BPPRed = (int)_RAWParameters.BPP2;
                BPPGreen = (int)_RAWParameters.BPP3;
                BPPBlue = (int)_RAWParameters.BPP4;
                MaskAlpha = _RAWParameters.Mask1;
                MaskRed = _RAWParameters.Mask2;
                MaskGreen = _RAWParameters.Mask3;
                MaskBlue = _RAWParameters.Mask4;
            }
            else
            {
                BPPAlpha = (int)_RAWParameters.BPP1;
                BPPRed = (int)_RAWParameters.BPP2;
                BPPGreen = (int)_RAWParameters.BPP3;
                BPPBlue = (int)_RAWParameters.BPP4;
                MaskAlpha = _RAWParameters.Mask1;
                MaskRed = _RAWParameters.Mask2;
                MaskGreen = _RAWParameters.Mask3;
                MaskBlue = _RAWParameters.Mask4;
            }

            UpdateStartBits();
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            //throw new Exception("The method or operation is not implemented.");

            Bitmap tempImage = new Bitmap(_RAWParameters.Width, _RAWParameters.Height);

            //get the palette first
            byte[] paletteData = ReadDataFromFile(_FilePath, 84, 512);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData, IgnoreAlpha);
            Bitmap tempPalette = paletteChannelData.CreateImage();

            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
            }

            byte[] rawData = ReadDataFromFile(_FilePath, 596, (_RAWParameters.Width * _RAWParameters.Height));
            //rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
            ImageData currentChannelData = new ImageData(this, _RAWParameters.Width, _RAWParameters.Height);
            currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0, IgnoreAlpha);
            //tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

            tempImage = currentChannelData.CreateImage();

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < _RAWParameters.Height; y++)
            {
                for (int x = 0; x < _RAWParameters.Width; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }

            return finalImage;
        }
    }
}
