// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class SR2PS2IndexedRAWImage : SegmentedRAWImage
    {
        public SR2PS2IndexedRAWImage(string path, RAWParameters parms, SegmentedRAWParameters segParms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _SegmentedRAWParameters = segParms;
            _TypeDescription = "Soul Reaver 2 (Playstation 2) RAW Bitmap";
            _FormatDescription = "Uncompressed Twiddled Indexed Image";

            _DataOffset = 144;
            BPPAlpha = (int)_RAWParameters.BPP1;
            BPPRed = (int)_RAWParameters.BPP2;
            BPPGreen = (int)_RAWParameters.BPP3;
            BPPBlue = (int)_RAWParameters.BPP4;
            MaskAlpha = _RAWParameters.Mask1;
            MaskRed = _RAWParameters.Mask2;
            MaskGreen = _RAWParameters.Mask3;
            MaskBlue = _RAWParameters.Mask4;

            UpdateStartBits();
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            //throw new Exception("The method or operation is not implemented.");
            Bitmap tempImage = new Bitmap(_RAWParameters.Width, _RAWParameters.Height);

            int subHorizontalCount = (ushort)(_RAWParameters.Width / 256);
            if ((subHorizontalCount * 256) < _RAWParameters.Width)
            {
                subHorizontalCount++;
            }
            //int subHorizontalCount = (ushort)(mImageWidth / 128);

            int[] subImageWidths = new int[32];
            int[] subImageHeights = new int[32];
            int imageCounter = 0;
            long positionCounter = 656;
            int vOffset = 0;
            ImageTransformer iTrans = new ImageTransformer();

            //get the palette first
            byte[] paletteData = ReadDataFromFile(_FilePath, 144, 512);
            ImageData paletteChannelData = new ImageData(this, 256, 1);
            paletteChannelData.GetChannelsFromRawBytes(paletteData, IgnoreAlpha);
            Bitmap tempPalette = paletteChannelData.CreateImage();

            //re-arrange the palette colours to match how the PS2 stores them
            Bitmap palette = new Bitmap(tempPalette);
            int[] clutIndex = BuildIDTex8ClutIndex();
            for (int i = 0; i < 256; i++)
            {
                palette.SetPixel(i, 0, tempPalette.GetPixel(clutIndex[i], 0));
            }

            //pretend we're dead^W a series of 8-bit images
            int rowCount = 0;
            do
            {
                int hOffset = 0;
                for (int j = 0; j < subHorizontalCount; j++)
                {
                    subImageWidths[imageCounter] = (ushort)Math.Min(256, _RAWParameters.Width - hOffset);
                    long currentDataLength = ReadUInt32FromFile(_FilePath, positionCounter) * 16;

                    subImageHeights[imageCounter] = (int)(currentDataLength / (subImageWidths[imageCounter]));

                    long subImageOffset = positionCounter + 16;

                    positionCounter += 16 + currentDataLength;

                    byte[] rawData = ReadDataFromFile(_FilePath, subImageOffset, currentDataLength);
                    rawData = unPS2DefianceSwizzle(rawData, subImageWidths[imageCounter], subImageHeights[imageCounter], 8);
                    ImageData currentChannelData = new ImageData(this, subImageWidths[imageCounter], subImageHeights[imageCounter]);
                    currentChannelData.GetChannelsFromRawBytes(rawData, 8, 0, 8, 8, 8, 0, 0, 0, 0, IgnoreAlpha);
                    tempImage = compositeImages(tempImage, currentChannelData.CreateImage(), hOffset, vOffset);

                    hOffset += subImageWidths[imageCounter];
                    imageCounter++;
                }
                //vOffset += subImageHeights[imageCounter - 1];
                vOffset += subImageHeights[(rowCount * subHorizontalCount)];
                rowCount++;
            } while (vOffset < _RAWParameters.Height);

            //use the resulting image as an index into the palette
            Bitmap finalImage = new Bitmap(tempImage);
            for (int y = 0; y < _RAWParameters.Height; y++)
            {
                for (int x = 0; x < _RAWParameters.Width; x++)
                {
                    //all of the channels in the tempImage except Alpha will have the index #. I am picking
                    //green arbitrarily.
                    finalImage.SetPixel(x, y, palette.GetPixel(tempImage.GetPixel(x, y).G, 0));
                }
            }

           return finalImage;
            
            
            
            
            //byte[] imageBytes = ReadBytes(_DataOffset, Width * Height * 2);
            //imageBytes = BLPS2.PS2ImageData.UnSwizzleImageBytes(imageBytes, Width, Height, BPPImage);  // 8 = 8BPP for the main image data
            //imageBytes = BLPS2.PS2ImageData.RGB565SwapRandB(imageBytes);

            //Bitmap image = CreateBitmap(Width, Height, Stride, PixelFormat.Format16bppRgb565, imageBytes);

            //return image;
        }
    }
}
