// BenLincoln.TheLostWorlds.CDImages
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDImages.

// BenLincoln.TheLostWorlds.CDImages is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDImages is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDImages (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using BLD = BenLincoln.Data;
using BLPS2 = BenLincoln.Playstation2;

namespace BenLincoln.TheLostWorlds.CDImages
{
    public class TRLPSPARGBRAWImage : RAWImage
    {
        protected PixelFormat _PixelFormat;

        public TRLPSPARGBRAWImage(string path, RAWParameters parms)
        {
            _FilePath = path;
            _RAWParameters = parms;
            _TypeDescription = "Tomb Raider: Legend (PSP) RAW Bitmap";

            _DataOffset = 128;
            //these values are hardcoded because the header information is incorrect in this filetype
            //settings are based on the file type (1 or 2)
            //imageBitsPerPixel = (ushort)tBPP;
            _RAWParameters.BPPImage = 16;

            if (parms.TypeFlags.Flags1 == 1)
            {
                _BPPAlpha = (ushort)0;
                _BPPRed = (ushort)5;
                _BPPGreen = (ushort)6;
                _BPPBlue = (ushort)5;
                _MaskAlpha = 0x00000000;
                _MaskBlue = 0x0000F800;
                _MaskGreen = 0x000007E0;
                _MaskRed = 0x0000001F;
                _FormatDescription = "Uncompressed Twiddled RGB 565 Image";
            }
            else
            {
                _BPPAlpha = (ushort)4;
                _BPPRed = (ushort)4;
                _BPPGreen = (ushort)4;
                _BPPBlue = (ushort)4;
                _MaskAlpha = 0x0000F000;
                _MaskBlue = 0x00000F00;
                _MaskGreen = 0x000000F0;
                _MaskRed = 0x0000000F;
                _FormatDescription = "Uncompressed Twiddled ARGB 4444 Image";
            }
        }

        public override System.Drawing.Bitmap GetBitmap()
        {
            byte[] imageBytes = ReadBytes(_DataOffset, Width * Height * 2);
            imageBytes = BLPS2.PS2ImageData.UnSwizzleImageBytes(imageBytes, Width, Height, BPPImage);  // 8 = 8BPP for the main image data
            imageBytes = BLPS2.PS2ImageData.RGB565SwapRandB(imageBytes);

            Bitmap image = CreateBitmap(Width, Height, Stride, PixelFormat.Format16bppRgb565, imageBytes);

            return image;
        }
    }
}
