// Soul Reaver 2 PC Trainer
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of Soul Reaver 2 PC Trainer.

// Soul Reaver 2 PC Trainer is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// Soul Reaver 2 PC Trainer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Soul Reaver 2 PC Trainer (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using BenLincoln.ActiveMemory;
using BenLincoln.HotKeys;

namespace SoulReaver2PCTrainer
{
    public partial class frmMain : Form
    {
        #region Enums

        protected enum Game
        {
            SR2PCEnglish102
        }

        #endregion

        #region Structs

        [StructLayout(LayoutKind.Explicit)]
        protected struct FloatAsBytes
        {
            [FieldOffset(0)]
            public byte ByteValue0;

            [FieldOffset(1)]
            public byte ByteValue1;

            [FieldOffset(2)]
            public byte ByteValue2;

            [FieldOffset(3)]
            public byte ByteValue3;

            [FieldOffset(0)]
            public float FloatValue;
        }

        #endregion

        protected MemoryFragment[] _ActionFreezeFrags;

        protected MemoryFragment _ActionStateFrozenFrag;

        protected MemoryFragment[] _ActionUnfreezeFrags;

        protected MemoryFragment[] _CameraFollowFrags;

        protected MemoryFragment[] _CameraFreeFrags;

        protected int _CurrentMenu;

        protected MemoryFragment[] _DebugMenuOffFrags;

        protected MemoryFragment[] _DebugMenuOnFrags;

        protected MemoryFragment _DebugMenuStateOnFrag;

        protected MemoryFragment[] _DebugMenuSwitchFrags;

        protected MemoryFragment[] _GameIDFrags;

        protected HotKey _hkActionToggle;

        protected HotKey _hkCameraFollow;

        protected HotKey _hkCameraFree;

        protected HotKey _hkDebugMenuNextMenu;

        protected HotKey _hkDebugMenuPreviousMenu;

        protected HotKey _hkDebugMenuToggle;

        protected HotKey _hkHUDToggle;

        protected HotKey _hkProjectionDistanceDecrement;

        protected HotKey _hkProjectionDistanceIncrement;

        protected MemoryFragment[] _HUDDrawFrags;

        protected MemoryFragment[] _HUDHideFrags;

        protected MemoryFragment _HUDStateHiddenFrag;

        protected string _ProcessName;

        protected MemoryFragment _ProjectionDistanceFrag;

        public frmMain()
        {
            this.InitializeComponent();
            this._ProcessName = "sr2";
            this._GameIDFrags = this.GetGameIDFrags(Game.SR2PCEnglish102);
            this._DebugMenuOnFrags = this.GetDebugMenuOnFrags(Game.SR2PCEnglish102);
            this._DebugMenuOffFrags = this.GetDebugMenuOffFrags(Game.SR2PCEnglish102);
            this._DebugMenuSwitchFrags = this.GetMenuSwitchFrags(Game.SR2PCEnglish102);
            this._CameraFreeFrags = this.GetCameraFreeFrags(Game.SR2PCEnglish102);
            this._CameraFollowFrags = this.GetCameraFollowFrags(Game.SR2PCEnglish102);
            this._ActionFreezeFrags = this.GetActionFreezeFrags(Game.SR2PCEnglish102);
            this._ActionUnfreezeFrags = this.GetActionUnfreezeFrags(Game.SR2PCEnglish102);
            this._HUDHideFrags = this.GetHUDHideFrags(Game.SR2PCEnglish102);
            this._HUDDrawFrags = this.GetHUDDrawFrags(Game.SR2PCEnglish102);
            this._DebugMenuStateOnFrag = this.GetDebugMenuStateOnFrag(Game.SR2PCEnglish102);
            this._ActionStateFrozenFrag = this.GetActionStateFrozenFrag(Game.SR2PCEnglish102);
            this._HUDStateHiddenFrag = this.GetHUDStateHiddenFrag(Game.SR2PCEnglish102);
            this._ProjectionDistanceFrag = this.GetProjectionDistanceFrag(Game.SR2PCEnglish102);
            this._CurrentMenu = 0;
            txtControls.Text =
                "F2\t= Toggle Debug Menu On/Off" + Environment.NewLine +
                "F3/F4\t= Previous/Next Debug Menu" + Environment.NewLine +
                "F5\t= Toggle In-Game Action On/Off" + Environment.NewLine +
                "F6\t= Toggle HUD On/Off" + Environment.NewLine +
                "F7\t= Free-Roaming Camera" + Environment.NewLine +
                "F8\t= Follow-Cam (Standard Camera)" + Environment.NewLine +
                "[Ctrl] +/-\t= Alter Projection Distance";
            this.RegisterHotKeys();
        }

        protected void ChangeProjectionDistance(float changeAmount)
        {
            byte[] numArray = this.ReadFrag(this._ProjectionDistanceFrag);
            if (numArray != null)
            {
                float single = BitConverter.ToSingle(numArray, 0);
                single = single + changeAmount;
                if (single > 3.402823E+38f)
                {
                    single = 3.402823E+38f;
                }
                if (single < -3.402823E+38f)
                {
                    single = -3.402823E+38f;
                }
                FloatAsBytes newFloatAsBytes = this.GetNewFloatAsBytes();
                newFloatAsBytes.FloatValue = single;
                byte[] byteValue0 = new byte[4];
                byteValue0[0] = newFloatAsBytes.ByteValue0;
                byteValue0[1] = newFloatAsBytes.ByteValue1;
                byteValue0[2] = newFloatAsBytes.ByteValue2;
                byteValue0[3] = newFloatAsBytes.ByteValue3;
                this._ProjectionDistanceFrag = new MemoryFragment(this._ProjectionDistanceFrag.Address, byteValue0);
                this.WriteFrag(this._ProjectionDistanceFrag);
                this.UpdateStatus(string.Concat("Set projection distance to ", single.ToString()));
                return;
            }
            else
            {
                this.UpdateStatus("Unable to change projection distance");
                return;
            }
        }

        protected MemoryFragment[] GetActionFreezeFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            memoryFragment[0] = new MemoryFragment(7351800, "00002000");
            return memoryFragment;
        }

        protected MemoryFragment GetActionStateFrozenFrag(Game game)
        {
            return new MemoryFragment(7351800, "00002000");
        }

        protected MemoryFragment[] GetActionUnfreezeFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            memoryFragment[0] = new MemoryFragment(7351800, "00100000");
            return memoryFragment;
        }

        protected MemoryFragment[] GetCameraFollowFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            //memoryFragment[0] = new MemoryFragment(7558708, "07");
            memoryFragment[0] = new MemoryFragment(7558708, "00");
            return memoryFragment;
        }

        protected MemoryFragment[] GetCameraFreeFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            //memoryFragment[0] = new MemoryFragment(7558708, "00");
            memoryFragment[0] = new MemoryFragment(7558708, "07");
            return memoryFragment;
        }

        protected MemoryFragment[] GetDebugMenuOffFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[2];
            memoryFragment[0] = new MemoryFragment(4844034, "E8B8270100");
            memoryFragment[1] = new MemoryFragment(4919750, "C705C452730000007043");
            return memoryFragment;
        }

        protected MemoryFragment[] GetDebugMenuOnFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[5];
            memoryFragment[0] = new MemoryFragment(4844034, "E829010000");
            memoryFragment[1] = new MemoryFragment(4844338, "C705C452730000007043C3");
            memoryFragment[2] = new MemoryFragment(4845451, "E9A2FBFFFF");
            memoryFragment[3] = new MemoryFragment(7351830, "04");
            memoryFragment[4] = new MemoryFragment(7557830, "D0");
            return memoryFragment;
        }

        protected MemoryFragment GetDebugMenuStateOnFrag(Game game)
        {
            return new MemoryFragment(7351830, "04");
        }

        protected MemoryFragment[] GetGameIDFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            memoryFragment[0] = new MemoryFragment(5150808, "31");
            return memoryFragment;
        }

        protected MemoryFragment[] GetHUDDrawFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            memoryFragment[0] = new MemoryFragment(4917709, "E8BE1C0000");
            return memoryFragment;
        }

        protected MemoryFragment[] GetHUDHideFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[1];
            memoryFragment[0] = new MemoryFragment(4917709, "9090909090");
            return memoryFragment;
        }

        protected MemoryFragment GetHUDStateHiddenFrag(Game game)
        {
            return new MemoryFragment(4917709, "9090909090");
        }

        protected MemoryFragment[] GetMenuSwitchFrags(Game game)
        {
            MemoryFragment[] memoryFragment = new MemoryFragment[10];
            memoryFragment[0] = new MemoryFragment(5719200, "10385700");
            memoryFragment[1] = new MemoryFragment(5719200, "D0385700");
            memoryFragment[2] = new MemoryFragment(5719200, "48395700");
            memoryFragment[3] = new MemoryFragment(5719200, "A8395700");
            memoryFragment[4] = new MemoryFragment(5719200, "503A5700");
            memoryFragment[5] = new MemoryFragment(5719200, "103B5700");
            memoryFragment[6] = new MemoryFragment(5719200, "903C5700");
            memoryFragment[7] = new MemoryFragment(5719200, "D03E5700");
            memoryFragment[8] = new MemoryFragment(5719200, "30425700");
            memoryFragment[9] = new MemoryFragment(5719200, "A8425700");
            return memoryFragment;
        }

        protected FloatAsBytes GetNewFloatAsBytes()
        {
            FloatAsBytes floatAsByte;
            floatAsByte.FloatValue = 0f;
            floatAsByte.ByteValue0 = 0;
            floatAsByte.ByteValue1 = 0;
            floatAsByte.ByteValue2 = 0;
            floatAsByte.ByteValue3 = 0;
            return floatAsByte;
        }

        protected MemoryFragment GetProjectionDistanceFrag(Game game)
        {
            return new MemoryFragment(7558648, "0000A043");
        }

        protected uint IdentifyGameProcess()
        {
            uint processIDByName = MemoryAccessor.GetProcessIDByName(this._ProcessName);
            MemoryFragment[] memoryFragmentArray = this._GameIDFrags;
            for (int i = 0; i < (int)memoryFragmentArray.Length; i++)
            {
                MemoryFragment memoryFragment = memoryFragmentArray[i];
                try
                {
                    byte[] numArray = MemoryAccessor.ReadMemory(processIDByName, memoryFragment.Address, (uint)(memoryFragment.Values.GetUpperBound(0) + 1));
                    bool flag = true;
                    for (int j = 0; j < numArray.GetUpperBound(0) + 1; j++)
                    {
                        if (numArray[j] != memoryFragment.Values[j])
                        {
                            flag = false;
                        }
                    }
                    if (!flag)
                    {
                    }
                    else
                    {
                        uint num = processIDByName;
                        return num;
                    }
                }
                catch (Exception exception)
                {
                }
            }
            throw new UnsupportedGameException(string.Concat("The version of '", this._ProcessName, "' running does not match any fingerprints."));
        }

        protected void NotifyWrongVersion()
        {
            this.UpdateStatus("The version of Soul Reaver 2 currently running is not supported.");
        }

        protected void OnHKActionToggle(object sender, EventArgs e)
        {
            if (this._ActionStateFrozenFrag.Equals(this.ReadFrag(this._ActionStateFrozenFrag)))
            {
                this.WriteFrags(this._ActionUnfreezeFrags);
                this.UpdateStatus("Unfreeze Action");
                return;
            }
            else
            {
                this.WriteFrags(this._ActionFreezeFrags);
                this.UpdateStatus("Freeze Action");
                return;
            }
        }

        protected void OnHKCameraFollow(object sender, EventArgs e)
        {
            this.WriteFrags(this._CameraFollowFrags);
            this.UpdateStatus("Follow Camera");
        }

        protected void OnHKCameraFree(object sender, EventArgs e)
        {
            this.WriteFrags(this._CameraFreeFrags);
            this.UpdateStatus("Free Camera");
        }

        protected void OnHKDebugMenuNextMenu(object sender, EventArgs e)
        {
            uint num;
            try
            {
                num = this.IdentifyGameProcess();
            }
            catch (Exception exception)
            {
                this.NotifyWrongVersion();
                return;
            }
            frmMain _frmMain = this;
            _frmMain._CurrentMenu = _frmMain._CurrentMenu + 1;
            if (this._CurrentMenu > this._DebugMenuSwitchFrags.GetUpperBound(0))
            {
                this._CurrentMenu = 0;
            }
            MemoryAccessor.WriteMemory(num, this._DebugMenuSwitchFrags[this._CurrentMenu].Address, this._DebugMenuSwitchFrags[this._CurrentMenu].Values);
            this.UpdateStatus("Next menu");
        }

        protected void OnHKDebugMenuPreviousMenu(object sender, EventArgs e)
        {
            uint num;
            try
            {
                num = this.IdentifyGameProcess();
            }
            catch (Exception exception)
            {
                this.NotifyWrongVersion();
                return;
            }
            frmMain _frmMain = this;
            _frmMain._CurrentMenu = _frmMain._CurrentMenu - 1;
            if (this._CurrentMenu < 0)
            {
                this._CurrentMenu = this._DebugMenuSwitchFrags.GetUpperBound(0);
            }
            MemoryAccessor.WriteMemory(num, this._DebugMenuSwitchFrags[this._CurrentMenu].Address, this._DebugMenuSwitchFrags[this._CurrentMenu].Values);
            this.UpdateStatus("Previous menu");
        }

        protected void OnHKDebugMenuToggle(object sender, EventArgs e)
        {
            if (this._DebugMenuStateOnFrag.Equals(this.ReadFrag(this._DebugMenuStateOnFrag)))
            {
                this.WriteFrags(this._DebugMenuOffFrags);
                this.UpdateStatus("Closed debug menu");
                return;
            }
            else
            {
                this.WriteFrags(this._DebugMenuOnFrags);
                this.UpdateStatus("Opened debug menu");
                return;
            }
        }

        protected void OnHKHUDToggle(object sender, EventArgs e)
        {
            if (this._HUDStateHiddenFrag.Equals(this.ReadFrag(this._HUDStateHiddenFrag)))
            {
                this.WriteFrags(this._HUDDrawFrags);
                this.UpdateStatus("Show HUD");
                return;
            }
            else
            {
                this.WriteFrags(this._HUDHideFrags);
                this.UpdateStatus("Hide HUD");
                return;
            }
        }

        protected void OnHKPDDecrement(object sender, EventArgs e)
        {
            this.ChangeProjectionDistance(-10f);
        }

        protected void OnHKPDIncrement(object sender, EventArgs e)
        {
            this.ChangeProjectionDistance(10f);
        }

        protected byte[] ReadFrag(MemoryFragment frag)
        {
            uint num;
            try
            {
                num = this.IdentifyGameProcess();
            }
            catch (Exception exception)
            {
                this.NotifyWrongVersion();
                byte[] numArray = null;
                return numArray;
            }
            return MemoryAccessor.ReadMemory(num, frag.Address, (uint)(frag.Values.GetUpperBound(0) + 1));
        }

        protected void RegisterHotKeys()
        {
            this._hkDebugMenuToggle = new HotKey(base.Handle, 6417, Keys.F2, HotKey.KeyModifiers.None, OnHKDebugMenuToggle);
            this._hkDebugMenuNextMenu = new HotKey(base.Handle, 6419, Keys.F4, HotKey.KeyModifiers.None, OnHKDebugMenuNextMenu);
            this._hkDebugMenuPreviousMenu = new HotKey(base.Handle, 6420, Keys.F3, HotKey.KeyModifiers.None, OnHKDebugMenuPreviousMenu);
            this._hkCameraFree = new HotKey(base.Handle, 6421, Keys.F7, HotKey.KeyModifiers.None, OnHKCameraFree);
            this._hkCameraFollow = new HotKey(base.Handle, 6422, Keys.F8, HotKey.KeyModifiers.None, OnHKCameraFollow);
            this._hkActionToggle = new HotKey(base.Handle, 6423, Keys.F5, HotKey.KeyModifiers.None, OnHKActionToggle);
            this._hkHUDToggle = new HotKey(base.Handle, 6424, Keys.F6, HotKey.KeyModifiers.None, OnHKHUDToggle);
            this._hkProjectionDistanceDecrement = new HotKey(base.Handle, 6425, Keys.Subtract, HotKey.KeyModifiers.Control, OnHKPDDecrement);
            this._hkProjectionDistanceIncrement = new HotKey(base.Handle, 6426, Keys.Add, HotKey.KeyModifiers.Control, OnHKPDIncrement);
        }

        protected void UpdateStatus(string status)
        {
            this.lblLastStatus.Text = string.Concat("Last Status: ", status);
            base.Invalidate();
        }

        protected void WriteFrag(MemoryFragment frag)
        {
            uint num;
            try
            {
                num = this.IdentifyGameProcess();
            }
            catch (Exception exception)
            {
                this.NotifyWrongVersion();
                return;
            }
            MemoryAccessor.WriteMemory(num, frag.Address, frag.Values);
        }

        protected void WriteFrags(MemoryFragment[] frags)
        {
            uint num;
            try
            {
                num = this.IdentifyGameProcess();
            }
            catch (Exception exception)
            {
                this.NotifyWrongVersion();
                return;
            }
            MemoryFragment[] memoryFragmentArray = frags;
            for (int i = 0; i < (int)memoryFragmentArray.Length; i++)
            {
                MemoryFragment memoryFragment = memoryFragmentArray[i];
                MemoryAccessor.WriteMemory(num, memoryFragment.Address, memoryFragment.Values);
            }
        }
    }
}